// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// base class for building autonomous AR.Drone 2.0 applications upon
// =================================================================

#include "appBase.h"
#include "navdata.h"
#include "hawaii/common/helpers.h"
#include "hawaii/common/ostream.h"
#include "hawaii/common/error.h"
#include <ardrone_autonomy/CamSelect.h>
#include <sensor_msgs/image_encodings.h>
#include <std_msgs/Empty.h>
#include <SDL/SDL_keysym.h>

// constructor to optionally activate the drone's bottom instead of front camera
DroneAppBase::DroneAppBase( DroneCamera droneCameraArg ) :
	
	// parameters
	droneCamera( droneCameraArg ),
	
	// read-only accessors for most recently received messages
	engaged(        engagedReal        ),
	cmdVelDelta(    cmdVelDeltaReal    ),
	keyboardKeys(   keyboardKeysReal   ),
	gamepadButtons( gamepadButtonsReal ),
	gamepadAnalog(  gamepadAnalogReal  ),
	imageFrontPtr(  imageFrontPtrReal  ),
	imageBottomPtr( imageBottomPtrReal ),
	navdata(        navdataReal        ),
	 
	// closed-loop PID controllers for each of the drone's degrees of freedom
	// developer note: parameters initially tuned according to "en.wikipedia.org/wiki/Ziegler%E2%80%93Nichols_method" 
	//                 using the PD-rule, i.e. kP = 0.8 * kU, kI = 0, kD = 0.1 * kP * TU
	controllerLongitudinal(  0.32, 0.0, 0.16, 1.0 ), // kU =  0.4, TU = 5.0
	controllerLateral(       0.32, 0.0, 0.16, 1.0 ), // kU =  0.4, TU = 5.0
	controllerVertical(      5.20, 0.0, 1.11, 1.0 ), // kU =  6.5, TU = 1.7
	controllerYaw(          13.60, 0.0, 1.14, 1.0 ), // kU = 17.0, TU = 0.6
	
	
	// base connection to ROS
	initializerROS( "droneAppAutonomous" ),
	spinner( 16 ), // use at least the number of callbacks
	nodeHandle(),
	
	// inbound communication with manual flight application
	subscriberAppEngaged(        nodeHandle.subscribe< std_msgs::Bool              >( "droneApp/engaged",        1, &DroneAppBase::callbackAppEngaged,        this ) ),
	subscriberAppCmdVelDelta(    nodeHandle.subscribe< geometry_msgs::Twist        >( "droneApp/cmd_velDelta",   1, &DroneAppBase::callbackAppCmdVelDelta,    this ) ),
	subscriberAppKeyboardKeys(   nodeHandle.subscribe< std_msgs::UInt8MultiArray   >( "droneApp/keyboardKeys",   1, &DroneAppBase::callbackAppKeyboardKeys,   this ) ),
	subscriberAppGamepadButtons( nodeHandle.subscribe< std_msgs::UInt8MultiArray   >( "droneApp/gamepadButtons", 1, &DroneAppBase::callbackAppGamepadButtons, this ) ),
	subscriberAppGamepadAnalog(  nodeHandle.subscribe< std_msgs::Float32MultiArray >( "droneApp/gamepadAnalog",  1, &DroneAppBase::callbackAppGamepadAnalog,  this ) ),
	timerManual( false ),
	
	// outbound communication with manual flight application
	publisherAppCmdVel(  nodeHandle.advertise< geometry_msgs::Twist >( "droneApp/cmd_vel", 1 ) ),
	publisherAppTakeoff( nodeHandle.advertise< std_msgs::Empty      >( "droneApp/takeoff", 1 ) ),
	publisherAppLand(    nodeHandle.advertise< std_msgs::Empty      >( "droneApp/land",    1 ) ),
	clientAppFlightAnim( nodeHandle.serviceClient< ardrone_autonomy::FlightAnim >( "droneApp/setflightanimation" ) ),
	
	// inbound communication directly with drone via "ardrone_autonomy"
	subscriberDroneImageFront(  nodeHandle.subscribe( "ardrone/front/image_raw",  1, &DroneAppBase::callbackDroneImageFront,  this ) ),
	subscriberDroneImageBottom( nodeHandle.subscribe( "ardrone/bottom/image_raw", 1, &DroneAppBase::callbackDroneImageBottom, this ) ),
	subscriberDroneCameraInfo(  nodeHandle.subscribe( "ardrone/camera_info",      1, &DroneAppBase::callbackDroneCameraInfo,  this ) ),
	subscriberDroneNavdata(     nodeHandle.subscribe( "ardrone/navdata",          1, &DroneAppBase::callbackDroneNavdata,     this ) ),
	timerDrone( false ),
	
	// outbound communication directly with drone via "ardrone_autonomy"
	clientDroneCameraChannel( nodeHandle.serviceClient< ardrone_autonomy::CamSelect  >( "ardrone/setcamchannel"   ) ),
	clientDroneLEDAnimation(  nodeHandle.serviceClient< ardrone_autonomy::LedAnim    >( "ardrone/setledanimation" ) ) {
	
	// initialize most recently received messages with default values
	this->engagedReal = false ;
	this->cmdVelDeltaReal = geometry_msgs::Twist() ;
	this->keyboardKeysReal.first.lock()   ; this->keyboardKeysReal.second.resize(   SDLK_LAST, 0    ) ; this->keyboardKeysReal.first.unlock()   ;
	this->gamepadButtonsReal.first.lock() ; this->gamepadButtonsReal.second.resize( 9,         0    ) ; this->gamepadButtonsReal.first.unlock() ;
	this->gamepadAnalogReal.first.lock()  ; this->gamepadAnalogReal.second.resize(  6,         0.0f ) ; this->gamepadAnalogReal.first.unlock()  ;
	(void)this->imageFrontPtrReal  ; // dev. note: This may well be a "nullptr".
	(void)this->imageBottomPtrReal ; // dev. note: This may well be a "nullptr".
	(void)this->navdataReal        ; // dev. note: Default c'tor is fine.
	
	// block until camera calibration parameters are valid before running derived class's c'tor
	while( ( this->droneCamera == droneCameraFront  && !this->undistorterFront.initialized()  )
	    || ( this->droneCamera == droneCameraBottom && !this->undistorterBottom.initialized() ) ) {
		hawaii::cout << "INFO: waiting for undistorter to be initialized..." << hawaii::endl ;
		ros::spinOnce() ;
		boost::this_thread::sleep( boost::posix_time::milliseconds( 100 ) ) ;
	}
}

// destructor
DroneAppBase::~DroneAppBase() {
	this->stop() ;
}

// run the drone application, wait for it to be finished or stop it immediately
void DroneAppBase::start() {
	
	// start ROS processing
	this->spinner.start() ;
}
void DroneAppBase::wait() {
	
	// loop until "arDroneManual" has ended
	size_t itersDead = 0 ;
	while( true ) {
		
		// also shut down if "arDroneManual" has not even started after 50 iterations
		if( !this->timerManual.isRunning() ) {
			if( ++itersDead >= 20 ) {
				hawaii::cerr << "ERROR: \"arDroneManual\" does not seem to be running, shutting down" << hawaii::endl ;
				break ;
			}
		}
		
		// shut down if last message from "arDroneManual" was more than 5 seconds ago
		else if( this->timerManual.toc() > 2000.0 ) {
			hawaii::cerr << "ERROR: \"arDroneManual\" has been stopped, shutting down" << hawaii::endl ;
			break ;
		}
		
		// wait 100ms per iteration
		boost::this_thread::sleep( boost::posix_time::milliseconds( 100 ) ) ;
	}
	
	// stop ROS processing
	this->spinner.stop() ;
}
void DroneAppBase::stop() {
	
	// stop ROS processing
	this->spinner.stop() ;
}

// automatically dis-/enable closed-loop PID controllers' integrators 
void DroneAppBase::controllersIntegratorAutoSet() {
	
	// enable controllers' integrators only if flying and if corresponding command has not been altered
	if( this->navdata.state == navdataStateFlying && this->cmdVelDelta.linear.x  == 0.0 ) { this->controllerLongitudinal.integratorEnable() ; } else { this->controllerLongitudinal.integratorDisable() ; } ;
	if( this->navdata.state == navdataStateFlying && this->cmdVelDelta.linear.y  == 0.0 ) { this->controllerLateral.integratorEnable()      ; } else { this->controllerLateral.integratorDisable()      ; } ;
	if( this->navdata.state == navdataStateFlying && this->cmdVelDelta.linear.z  == 0.0 ) { this->controllerVertical.integratorEnable()     ; } else { this->controllerVertical.integratorDisable()     ; } ;
	if( this->navdata.state == navdataStateFlying && this->cmdVelDelta.angular.z == 0.0 ) { this->controllerYaw.integratorEnable()          ; } else { this->controllerYaw.integratorDisable()          ; } ;
}

// continuous flight control
void DroneAppBase::sendFlightCommands( FlightCommands flightCommands ) {
	
	// check given commands for "NAN"
	if( isnan( flightCommands.longitudinal ) ) { flightCommands.longitudinal = 0.0 ; hawaii::cerr << "WARNING: replaced \"NAN\" longitudinal flight command by zero before sending." << hawaii::endl ; }
	if( isnan( flightCommands.lateral      ) ) { flightCommands.lateral      = 0.0 ; hawaii::cerr << "WARNING: replaced \"NAN\" lateral flight command by zero before sending."      << hawaii::endl ; }
	if( isnan( flightCommands.vertical     ) ) { flightCommands.vertical     = 0.0 ; hawaii::cerr << "WARNING: replaced \"NAN\" vertical flight command by zero before sending."     << hawaii::endl ; }
	if( isnan( flightCommands.yaw          ) ) { flightCommands.yaw          = 0.0 ; hawaii::cerr << "WARNING: replaced \"NAN\" yaw flight command by zero before sending."          << hawaii::endl ; }
	
	// limit and send commands
	geometry_msgs::Twist cmdVel ;
	cmdVel.linear.x  = std::min( 1.0, std::max( -1.0, flightCommands.longitudinal ) ) ;
	cmdVel.linear.y  = std::min( 1.0, std::max( -1.0, flightCommands.lateral      ) ) ;
	cmdVel.linear.z  = std::min( 1.0, std::max( -1.0, flightCommands.vertical     ) ) ;
	cmdVel.angular.x =                                0.0                             ;
	cmdVel.angular.y =                                0.0                             ;
	cmdVel.angular.z = std::min( 1.0, std::max( -1.0, flightCommands.yaw          ) ) ;
	this->publisherAppCmdVel.publish( cmdVel ) ;
}

// hover in-place
void DroneAppBase::sendFlightCommandsHover() {
	
	// activate on-board stabilization by sending all zeros
	geometry_msgs::Twist cmdVel ;
	cmdVel.linear.x  = 0.0 ;
	cmdVel.linear.y  = 0.0 ;
	cmdVel.linear.z  = 0.0 ;
	cmdVel.angular.x = 0.0 ;
	cmdVel.angular.y = 0.0 ;
	cmdVel.angular.z = 0.0 ;
	this->publisherAppCmdVel.publish( cmdVel ) ;
}

// discrete flight maneuvers
void DroneAppBase::sendTakeoff() { this->publisherAppTakeoff.publish( std_msgs::Empty() ) ; }
void DroneAppBase::sendLanding() { this->publisherAppLand.publish(    std_msgs::Empty() ) ; }
void DroneAppBase::sendFlightAnimation( ardrone_autonomy::FlightAnim& flightAnim ) {
	this->sendFlightAnimation( flightAnim.request,
	                           flightAnim.response ) ;
}
void DroneAppBase::sendFlightAnimation( const ardrone_autonomy::FlightAnimRequest&  flightAnimRequest,
                                              ardrone_autonomy::FlightAnimResponse& flightAnimResponse ) {
	this->clientAppFlightAnim.call( flightAnimRequest,
	                                flightAnimResponse ) ;
}

// position LED signals
void DroneAppBase::sendLEDAnimation( ardrone_autonomy::LedAnim& ledAnim ) {
	this->sendLEDAnimation( ledAnim.request,
	                        ledAnim.response ) ;
}
void DroneAppBase::sendLEDAnimation( const ardrone_autonomy::LedAnimRequest&  ledAnimRequest,
                                           ardrone_autonomy::LedAnimResponse& ledAnimResponse ) {
	this->clientDroneLEDAnimation.call( ledAnimRequest,
	                                    ledAnimResponse ) ;
}

// flag to indicate if autonomous control is in effect at all
void DroneAppBase::callbackAppEngaged( const std_msgs::Bool::ConstPtr& engagedPtr ) {
	
	// heartbeat for time-out detection
	this->timerManual.tic() ;
	
	// buffer this message and trigger derived application's processing
	this->engagedReal = engagedPtr->data ;
	this->processEngaged() ;
}

// modifications applied to autonomous control commands
void DroneAppBase::callbackAppCmdVelDelta( const geometry_msgs::Twist::ConstPtr& cmdVelDeltaPtr ) {
	
	// heartbeat for time-out detection
	this->timerManual.tic() ;
	
	// buffer this message and trigger derived application's processing
	this->cmdVelDeltaReal = *cmdVelDeltaPtr ;
	this->controllersIntegratorAutoSet() ;
	this->processCmdVelDelta() ;
}

// states of keyboard keys
void DroneAppBase::callbackAppKeyboardKeys( const std_msgs::UInt8MultiArray::ConstPtr& keyboardKeysPtr ) {
	
	// heartbeat for time-out detection
	this->timerManual.tic() ;
	
	// buffer this message and trigger derived application's processing
	this->keyboardKeysReal.first.lock() ;
	this->keyboardKeysReal.second = keyboardKeysPtr->data ;
	this->keyboardKeysReal.first.unlock() ;
	this->processKeyboardAndGamepad() ;
}

// states of gamepad buttons and digital cross
void DroneAppBase::callbackAppGamepadButtons( const std_msgs::UInt8MultiArray::ConstPtr& gamepadButtonsPtr ) {
	
	// heartbeat for time-out detection
	this->timerManual.tic() ;
	
	// buffer this message and trigger derived application's processing
	this->gamepadButtonsReal.first.lock() ;
	this->gamepadButtonsReal.second = gamepadButtonsPtr->data ;
	this->gamepadButtonsReal.first.unlock() ;
	this->processKeyboardAndGamepad() ;
}

// positions of gamepad analog sticks and triggers
void DroneAppBase::callbackAppGamepadAnalog( const std_msgs::Float32MultiArray::ConstPtr& gamepadAnalogPtr ) {
	
	// heartbeat for time-out detection
	this->timerManual.tic() ;
	
	// buffer this message and trigger derived application's processing
	this->gamepadAnalogReal.first.lock() ;
	this->gamepadAnalogReal.second = gamepadAnalogPtr->data ;
	this->gamepadAnalogReal.first.unlock() ;
	this->processKeyboardAndGamepad() ;
}

// image from front camera
void DroneAppBase::callbackDroneImageFront( const sensor_msgs::Image::ConstPtr& imageFrontPtr ) {
	
	// heartbeat for time-out detection
	this->timerDrone.tic() ;
	
	// if wrong camera, switch and return
	if( this->droneCamera != droneCameraFront ) {
		hawaii::cout << "INFO: switching to bottom camera" << hawaii::endl ;
		ardrone_autonomy::CamSelect camSelect ;
		camSelect.request.channel = 1 ;
		this->clientDroneCameraChannel.call( camSelect ) ;
		if( !camSelect.response.result ) { hawaii::cerr << "ERROR: failed to switch to bottom camera." << hawaii::endl ; }
		return ;
	}
	
	// only if corresponding undistorter is initialized...
	if( this->undistorterFront.initialized() ) {
		
		// convert image to OpenCV
		cv_bridge::CvImage::Ptr toBeUndistortedPtr = cv_bridge::toCvCopy( imageFrontPtr, sensor_msgs::image_encodings::BGR8 ) ;
		const cv::Mat distorted = toBeUndistortedPtr->image ;
		
		// undistort
		this->undistorterFront.CPU( distorted, toBeUndistortedPtr->image ) ;
		
		// buffer result and trigger derived application's processing
		this->imageFrontPtrReal = toBeUndistortedPtr ;
		this->processImageFront() ;
	}
}

// image from bottom camera
void DroneAppBase::callbackDroneImageBottom( const sensor_msgs::Image::ConstPtr& imageBottomPtr ) {
	
	// heartbeat for time-out detection
	this->timerDrone.tic() ;
	
	// if wrong camera, switch and return
	if( this->droneCamera != droneCameraBottom ) {
		hawaii::cout << "INFO: switching to front camera" << hawaii::endl ;
		ardrone_autonomy::CamSelect camSelect ;
		camSelect.request.channel = 0 ;
		this->clientDroneCameraChannel.call( camSelect ) ;
		if( !camSelect.response.result ) { hawaii::cerr << "ERROR: failed to switch to front camera." << hawaii::endl ; }
		return ;
	}
	
	// only if corresponding undistorter is initialized...
	if( this->undistorterBottom.initialized() ) {
		
		// undistortion not really necessary for bottom camera
		this->imageBottomPtrReal = cv_bridge::toCvCopy( imageBottomPtr, sensor_msgs::image_encodings::BGR8 ) ;
		
/*		// convert image to OpenCV
		cv_bridge::CvImage::Ptr toBeUndistortedPtr = cv_bridge::toCvCopy( imageBottomPtr, sensor_msgs::image_encodings::BGR8 ) ;
		const cv::Mat distorted = toBeUndistortedPtr->image ;
		
		// undistort
		this->undistorterBottom.CPU( distorted, toBeUndistortedPtr->image ) ;
		
		// buffer result and trigger derived application's processing
		this->imageBottomPtrReal = toBeUndistortedPtr ;
*/		this->processImageBottom() ;
	}
}

// parameters of currently active camera
void DroneAppBase::callbackDroneCameraInfo( const sensor_msgs::CameraInfo::ConstPtr& cameraInfoPtr ) {
	
	// heartbeat for time-out detection
	this->timerDrone.tic() ;
	
	// get active camera's running number
	DroneCamera droneCameraActive ;
	     if( cameraInfoPtr->header.frame_id == "ardrone_base_frontcam"  ) { droneCameraActive = droneCameraFront  ; }
	else if( cameraInfoPtr->header.frame_id == "ardrone_base_bottomcam" ) { droneCameraActive = droneCameraBottom ; }
	else { HAWAII_ERROR( "unknown camera" ) ; }
	
	// select corresponding undistorter
	toast2::stereo::Undistorter& undistorter( droneCameraActive == droneCameraFront ?
	                                          this->undistorterFront :
	                                          this->undistorterBottom                 ) ;
	
	// directly return if the undistorter is already initialized
	if( HAWAII_LIKELY( undistorter.initialized() ) ) { return ; }
	
	// check camera parameters
	if( cameraInfoPtr->K[0] <= 0.0
	 || cameraInfoPtr->K[2] <= 0.0
	 || cameraInfoPtr->K[4] <= 0.0
	 || cameraInfoPtr->K[5] <= 0.0 ) {
		hawaii::cerr << "ERROR: camera not yet calibrated" << hawaii::endl ;
		return ;
	}
	if( cameraInfoPtr->D.size() != 0
	 && cameraInfoPtr->D.size() != 5 ) {
		hawaii::cerr << "ERROR: unsupported lens distortion model" << hawaii::endl ;
		return ;
	}
	
	// initialize undistorter
	hawaii::cout << "INFO: initializing " << ( droneCameraActive == droneCameraFront ? "front" : "bottom" ) << " undistorter" << hawaii::endl ;
	toast2::stereo::Distortion distortion ;
	if( !cameraInfoPtr->D.empty() ) {
		distortion(0) = cameraInfoPtr->D[0] ;
		distortion(1) = cameraInfoPtr->D[1] ;
		distortion(2) = cameraInfoPtr->D[2] ;
		distortion(3) = cameraInfoPtr->D[3] ;
		distortion(4) = cameraInfoPtr->D[4] ;
	}
	toast2::stereo::Camera cameraSrc ;
	cameraSrc(0,0) = cameraInfoPtr->K[0] ;
	cameraSrc(0,1) = cameraInfoPtr->K[1] ;
	cameraSrc(0,2) = cameraInfoPtr->K[2] ;
	cameraSrc(1,0) = cameraInfoPtr->K[3] ;
	cameraSrc(1,1) = cameraInfoPtr->K[4] ;
	cameraSrc(1,2) = cameraInfoPtr->K[5] ;
	cameraSrc(2,0) = cameraInfoPtr->K[6] ;
	cameraSrc(2,1) = cameraInfoPtr->K[7] ;
	cameraSrc(2,2) = cameraInfoPtr->K[8] ;
	cv::Size sizeSrc ;
	sizeSrc.width  = cameraInfoPtr->width  ;
	sizeSrc.height = cameraInfoPtr->height ;
	undistorter.update( distortion, cameraSrc, sizeSrc, true ) ;
}

// combined sensor data
void DroneAppBase::callbackDroneNavdata( const ardrone_autonomy::Navdata::ConstPtr& navdataPtr ) {
	
	// heartbeat for time-out detection
	this->timerDrone.tic() ;
	
	// buffer this message and trigger derived application's processing
	this->navdataReal = *navdataPtr ;
	this->controllersIntegratorAutoSet() ;
	this->processNavdata() ;
}
