// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Authors: Benjamin Ranft (benjamin.ranft@web.de), Jacques Maylaender
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// follow a red line on the ground
// ===============================

#include "appFollowLine.h"
#include "lineDetection.h"
#include "gamepad.h"
#include "clozer.h"
#include "hawaii/common/helpers.h"
#include "hawaii/common/ostream.h"
#include <opencv2/imgproc/imgproc.hpp>
#include <SDL/SDL_keysym.h>
#include <SDL/SDL_joystick.h>

// constructor
DroneAppFollowLine::DroneAppFollowLine() :
	
	// activate bottom camera
	DroneAppBase( droneCameraBottom ),
	
	// forward speed
	speedLongitudinal( 0.05 ),
	
	// target height
	heightTarget( 1.5 ), // STUD_TODO high enough so that line is not blown away
	
	// target absolute yaw angle of drone
	yawTarget( NAN ),
	
	// relative lateral offset from drone to line
	lateralOffset( NAN ),
	
	// time when line was last seen and when last "navdata" were received
	timerLine(    false ),
	timerNavdata( false ),
	
	// trackers for smoothing yaw and lateral error
	smootherYaw(     0.06 ),
	smootherLateral( 0.06 ) {
	
	// copy measured critical gain and period from Ziegler-Nichols tuning of height control
	const double kPCritVertical =  6.5 ; // STUD_SET_NAN
	const double TCritVertical  =  1.7 ; // STUD_SET_NAN
	this->controllerVertical.kP = 0.8 * kPCritVertical                 ;
	this->controllerVertical.kI = 0.0 * kPCritVertical / TCritVertical ;
	this->controllerVertical.kD = 0.1 * kPCritVertical * TCritVertical ;
	
	// same as above for yaw control
	const double kPCritYaw = 12.0 ; // STUD_SET_NAN
	const double TCritYaw  =  0.5 ; // STUD_SET_NAN
	this->controllerYaw.kP = 0.8 * kPCritYaw            ;
	this->controllerYaw.kI = 0.0 * kPCritYaw / TCritYaw ;
	this->controllerYaw.kD = 0.1 * kPCritYaw * TCritYaw ;
	
	// same as above for lateral control
	const double kPCritLateral = 0.4 ;
	const double TCritLateral  = 5.0 ;
	this->controllerLateral.kP = 0.8 * kPCritLateral                ;
	this->controllerLateral.kI = 0.0 * kPCritLateral / TCritLateral ;
	this->controllerLateral.kD = 0.1 * kPCritLateral * TCritLateral ;
	
} // constructor "DroneAppFollowLine()"

// visual line detection
void DroneAppFollowLine::processImageBottom() {
	
	// buffer current image and navdata from the same point in time
	const cv::Mat                   image          = this->imageBottomPtr->image ;
	const ardrone_autonomy::Navdata navdataAtImage = this->navdata               ;
	
	// create a copy of the input image to draw visualizations on
	cv::Mat visu ;
	image.copyTo( visu ) ;
	
	// define center
	const double uC = image.cols * 0.50,
	             vC = image.rows * 0.75 ;
	
	// draw image center coordinate system
	cv::line( visu,
	          cv::Point( uC, 0             ),
	          cv::Point( uC, visu.rows - 1 ),
	          cv::Scalar( 0, 255, 0 ), 1     ) ;
	cv::line( visu,
	          cv::Point( 0,             vC ),
	          cv::Point( visu.cols - 1, vC ),
	          cv::Scalar( 0, 255, 0 ), 1     ) ;
	
	// 1) roll angle compensation 
	// ==========================
	
	// roll angle from world to drone in [rad]
	const double rollWorldDrone = navdataAtImage.rotX / 180.0 * CV_PI ;
	
	// roll-induced pixel offset from image center to point below drone
	const double deltaUF = STUD_FUNCTION_NAN( atan )( rollWorldDrone ) * this->focalLengthBottomU() ; // STUD_TODO
	
	// project the drone's forward axis onto the image directly under it
	cv::line( visu,
	          cv::Point( uC + deltaUF, 0             ),
	          cv::Point( uC + deltaUF, visu.rows - 1 ),
	          cv::Scalar( 255, 255, 0 ), 2             ) ;
	
	// 2) line detection
	// =================
	
	// call line detection function
	const Coeffs coeffs = detectLine( image ) ;
	
	// if line has been found... 
	if( isnan( coeffs[ 0 ] ) == false
	 && isnan( coeffs[ 1 ] ) == false
	 && isnan( coeffs[ 2 ] ) == false
	 && isnan( coeffs[ 3 ] ) == false ) {
		
		// reset timer
		this->timerLine.tic() ;
	
	// otherwise...
	} else {
		
		// show the visualization and return without further computations
		cv::imshow( "visu", visu ) ;
		cv::waitKey( 1 ) ;
		return ;
	}
	
	// 3) yaw control
	// ==============
	
	// image row to evaluate line model at
	const double v = 100 ; // STUD_TODO tune
	
	// "preview length" from row "v" to center
	const double l = vC - v ;
	
	// image column of line model at row "v"
	const double u = coeffs.u( v ) ; // STUD_SET_NAN
	
	// horizontal pixel offset from center to column "u"
	const double deltaU = u - uC ;
	
	// drawbar angle ("Deichselwinkel") for steering towards line point "(u,v)"
	const double drawbar = - atan( ( deltaU - deltaUF ) / STUD_NAN( l ) ) ; // STUD_TODO triangle's adjacent ("Ankathete")?
	
	// draw connection from image center to line point
	cv::line( visu,
	          cv::Point( u,            v  ),
	          cv::Point( uC + deltaUF, vC ),
	          cv::Scalar( 255, 0, 0 ), 3    ) ;
	
	// derivative of line at image row "v"
	const double du_dv = 3 * coeffs[ 0 ] * v * v + 2 * coeffs[ 1 ] * v + coeffs[ 2 ] ; // STUD_SET_NAN
	
	// angle from drone's forward axis to tangent at line point
	const double tangent = STUD_FUNCTION_NAN( atan )( du_dv ) ; // STUD_TODO
	
	// draw tangent to line model
	cv::line( visu,
	          cv::Point( u - du_dv * v,                 0             ),
	          cv::Point( u + du_dv * ( visu.rows - v ), visu.rows - 1 ),
	          cv::Scalar( 255, 0, 255 ), 2                             ) ;
	
	// update absolute target yaw angle of drone
	this->yawTarget = 0.2 * drawbar + 0.8 * tangent + navdataAtImage.rotZ * CV_PI / 180.0 ; // STUD_TODO tune weights of drawbar and tangent angle
	
	// 4) lateral control
	// ==================
	
	// height above ground
	const double height = navdataAtImage.altd != 0       // Only if the height sensor if on, ...
	                    ? navdataAtImage.altd / 1000.0   // ...use the measured height.
	                    : this->heightTarget           ; // Otherwise (because the drone is carried), assume target height.
	
	// roll angle from drone to line in drone and world coordinate system
	const double rollDroneLine = - atan( STUD_NAN( deltaU ) / this->focalLengthBottomU() ) ; // STUD_TODO
	const double rollWorldLine = rollWorldDrone + rollDroneLine ;
	
	// update relative lateral offset from drone to line
	this->lateralOffset = STUD_NAN( height ) * tan( rollWorldLine ) ; // STUD_TODO
	
	// show the visualization before exiting
	cv::imshow( "visu", visu ) ;
	cv::waitKey( 1 ) ;
	
} // method "processImageBottom()"

// flight control
void DroneAppFollowLine::processNavdata() {
	
	// hover in place if line is lost for too long
	const bool lost = this->timerLine.isRunning() == false
	               || this->timerLine.toc()       >= 500.0 ;
	if( lost == true ) {
		hawaii::cerr << "WARNING: line has been lost for too long, controlling height only..." << hawaii::endl ;
	}
		
	// commands to be sent, use constant forward speed
	FlightCommands commands ;
	if( lost == false ) {
		commands.longitudinal = this->speedLongitudinal ;
	}
	
	// control yaw angle
	if( lost                     == false
	 && isnan( this->yawTarget ) == false ) {
		
		// compute error between absolute target (older) and actual (current) yaw angle
		double errorYaw = this->yawTarget - this->navdata.rotZ * CV_PI / 180.0 ;
		
		// wrap around until error is within [-pi, pi)
		while( errorYaw >=   CV_PI ) { errorYaw -= 2 * CV_PI ; }
		while( errorYaw <  - CV_PI ) { errorYaw += 2 * CV_PI ; }
		
		// compute and smoothen resulting control output
		this->smootherYaw( this->controllerYaw( errorYaw ) ) ;
		commands.yaw = this->smootherYaw ;
	}
	
	// control lateral position
	if( lost                         == false
	 && isnan( this->lateralOffset ) == false ) {
		
		// get sideways speed to be used below
		const double speedLateral = this->navdata.vy / 1000.0 ;
		
		// compute and smoothen resulting control output
		this->smootherLateral( this->controllerLateral( this->lateralOffset, false ? - speedLateral : NAN ) ) ;
		commands.lateral = this->smootherLateral ;
		
		// update relative lateral offset by integrating speed
		if( this->timerNavdata.isRunning() ) {
			this->lateralOffset -= speedLateral * this->timerNavdata.tocTic() / 1000.0 ;
		} else {
			this->timerNavdata.tic() ;
		}
	}
	
	// control height
	commands.vertical = this->controllerVertical( this->heightTarget - this->navdata.altd / 1000.0 ) ;
	
	// send commands to drone
	this->sendFlightCommands( commands ) ;
	
} // method "processNavdata()"

// adjust forward speed with "+"/"-" keyboard keys or up/down on gamepad's digital cross
void DroneAppFollowLine::processKeyboardAndGamepad() {
	
	// flags to be set below
	bool increase = false,
	     decrease = false ;
	
	// evaluate keyboard
	this->keyboardKeys.first.lock_shared() ;
	increase |= this->keyboardKeys.second[ SDLK_PLUS     ] ;
	increase |= this->keyboardKeys.second[ SDLK_KP_PLUS  ] ;
	decrease |= this->keyboardKeys.second[ SDLK_MINUS    ] ;
	decrease |= this->keyboardKeys.second[ SDLK_KP_MINUS ] ;
	this->keyboardKeys.first.unlock_shared() ;
	
	// evaluate gamepad
	this->gamepadButtons.first.lock_shared() ;
	increase |= this->gamepadButtons.second[ gamepadDigitalCross ] & SDL_HAT_UP   ;
	decrease |= this->gamepadButtons.second[ gamepadDigitalCross ] & SDL_HAT_DOWN ;
	this->gamepadButtons.first.unlock_shared() ;
	
	// apply result
	if( increase ) {
		this->speedLongitudinal *= 1.005 ;
		this->speedLongitudinal = std::min( 0.200, this->speedLongitudinal ) ;
	}
	if( decrease ) {
		this->speedLongitudinal /= 1.005 ;
		this->speedLongitudinal = std::max( 0.001, this->speedLongitudinal ) ;
	}
	
	// print new range
	if( increase || decrease ) {
		hawaii::cout << "INFO: forward speed = " << this->speedLongitudinal << hawaii::endl ;
	}
	
} // method "processKeyboardAndGamepad()"
