// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Authors: Benjamin Ranft (benjamin.ranft@web.de), Jacques Maylaender
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// tuning of controller parameters with the Ziegler-Nichols method
// ===============================================================

#pragma once

#include "appBase.h"
#include "clozer.h"
#include "gamepad.h"
#include "hawaii/common/tracker.h"
#include "hawaii/common/ostream.h"
#include <SDL/SDL_keysym.h>
#include <SDL/SDL_joystick.h>

// derived from base drone application
class DroneAppZieglerNichols : public DroneAppBase {
	
	// gain of P-controller
	protected:
	double kP ;
	
	// tracker for smoothing height or yaw measurements
	protected:
	hawaii::Tracker1DPT1 actual ;
	
	// constructor
	public:
	DroneAppZieglerNichols() :
		
		// activate bottom camera
		DroneAppBase( droneCameraBottom ) {
		
		// initialize P-controller gain with a low (!) value: Its represents normalized control command per height or 
		//                                                    yaw error, so its unit is [-1.0, +1.0] / [m] or [rad].
		this->kP = 0.1 ; // STUD_SET_NAN
		
		// PT1 time constant of the tracker for smoothing height or yaw measurements
		this->actual.reset( 0.03 ) ;
		
		// later, after manual tuning of height control...
		
		// ...enter your measured critical gain and period
		const double kPCritVertical =  6.5 ; // STUD_SET_NAN
		const double TCritVertical  =  1.7 ; // STUD_SET_NAN
		
		// ...use Ziegler-Nichols equations to find parameters for integrated PID controller
		this->controllerVertical.kP = STUD_NAN( 0.800 ) * kPCritVertical                 ; // STUD_TODO enter factors from script
		this->controllerVertical.kI = STUD_NAN( 0.000 ) * kPCritVertical / TCritVertical ; // STUD_TODO
		this->controllerVertical.kD = STUD_NAN( 0.100 ) * kPCritVertical * TCritVertical ; // STUD_TODO
		
		// later, after manual tuning of heading control...
		
		// ...enter your measured critical gain and period
		const double kPCritYaw = 12.0 ; // STUD_SET_NAN
		const double TCritYaw  =  0.5 ; // STUD_SET_NAN
		
		// ...use Ziegler-Nichols equations to find parameters for integrated PID controller
		this->controllerYaw.kP = STUD_NAN( 0.800 ) * kPCritYaw            ; // STUD_TODO enter factors from script
		this->controllerYaw.kI = STUD_NAN( 0.000 ) * kPCritYaw / TCritYaw ; // STUD_TODO
		this->controllerYaw.kD = STUD_NAN( 0.100 ) * kPCritYaw * TCritYaw ; // STUD_TODO
	}
	
	// compute and send new flight commands each time a new measurement comes in
	protected:
	virtual void processNavdata() override {
		
		// commands to be sent to drone
		FlightCommands flightCommands ;
		
		// helper variables
		double target = NAN,
		       error  = NAN ;
		
		// two degrees of freedom ("Freiheitsgrade") to be tuned
		enum { vertical, yaw } mode = vertical ; // STUD_TODO tune both
		switch( mode ) {
			
			// height control
			case vertical : {
				
				// define target height in [m]
				target = 1.0 ; // STUD_SET_NAN
				
				// read actual height, find its unit with the "rqt_plot" tool and convert it to [m]
				this->actual( STUD_NAN( this->navdata.altd / 1000.0 ) ) ; // STUD_TODO
				
				// compute height error and vertical speed command
				error = target - this->actual ; // STUD_SET_NAN
				flightCommands.vertical = STUD_NAN( this->kP ) * error ; // STUD_TODO multiply with proportional control factor
				
				// only if Ziegler-Nichols tuning is finished, overwrite with integrated PID controller
				if( isnan( this->controllerVertical.kP ) == false
				 && isnan( this->controllerVertical.kI ) == false
				 && isnan( this->controllerVertical.kD ) == false ) {
					flightCommands.vertical = this->controllerVertical( error ) ;
				}
				
				// exit "switch" block
				break ;
			}
			
			// heading control
			case yaw : {
				
				// define target yaw in [rad]
				target = 0.0 ;
				
				// read actual yaw, find its unit with the "rqt_plot" tool and convert it to [rad]
				this->actual( STUD_NAN( this->navdata.rotZ / 180.0 * CV_PI ) ) ; // STUD_TODO
				
				// compute yaw error and yaw rate command 
				error = target - this->actual ;
				flightCommands.yaw = this->kP * error ;
				
				// only if Ziegler-Nichols tuning is finished, overwrite with integrated PID controller
				if( isnan( this->controllerYaw.kP ) == false
				 && isnan( this->controllerYaw.kI ) == false
				 && isnan( this->controllerYaw.kD ) == false ) {
					flightCommands.yaw = this->controllerYaw( error ) ;
				}
				
				// exit "switch" block
				break ;
			}
		}
		
		// send commands to drone
		this->sendFlightCommands( flightCommands ) ;
		
		// print all values
		hawaii::cout << "target           = " << target                  << "\n"
		             << "actual           = " << this->actual            << "\n"
		             << "error            = " << error                   << "\n"
		             << "kP               = " << this->kP                << "\n"
		             << "command vertical = " << flightCommands.vertical << "\n"
		             << "command yaw      = " << flightCommands.yaw      << "\n" << hawaii::endl ;
	}
	
	// adjust P-controller gain with "+"/"-" keyboard keys or up/down on gamepad's digital cross
	protected:
	virtual void processKeyboardAndGamepad() override {
		
		// flags to be set below
		bool increase = false,
		     decrease = false ;
		
		// evaluate keyboard
		this->keyboardKeys.first.lock_shared() ;
		increase |= this->keyboardKeys.second[ SDLK_PLUS     ] ;
		increase |= this->keyboardKeys.second[ SDLK_KP_PLUS  ] ;
		decrease |= this->keyboardKeys.second[ SDLK_MINUS    ] ;
		decrease |= this->keyboardKeys.second[ SDLK_KP_MINUS ] ;
		this->keyboardKeys.first.unlock_shared() ;
		
		// evaluate gamepad
		this->gamepadButtons.first.lock_shared() ;
		increase |= this->gamepadButtons.second[ gamepadDigitalCross ] & SDL_HAT_UP   ;
		decrease |= this->gamepadButtons.second[ gamepadDigitalCross ] & SDL_HAT_DOWN ;
		this->gamepadButtons.first.unlock_shared() ;
		
		// apply result
		if( increase ) { this->kP *= 1.005 ; }
		if( decrease ) { this->kP /= 1.005 ; }
	}
	
} ; // class "DroneAppZieglerNichols"
