// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// undistortion, rectification and cropping of raw camera images
// =============================================================

#include "toast2/stereo/undistortRectify.h"
#include "hawaii/GPU/pool.h"
#include "hawaii/common/partitionize.h"
#include "hawaii/common/error.h"
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/core/core_c.h>

// helpers only used here
namespace {
	
	// transform camera/projection matrix and LUTs such that focal lengths are equal
	template< typename Intrinsic >
	void equalizeFocalLengths( const Intrinsic  intrin,
	                           const cv::Mat    mapU,
	                           const cv::Mat    mapV,
	                                 Intrinsic& intrinEqFoc,
	                                 cv::Mat&   mapUEqFoc,
	                                 cv::Mat&   mapVEqFoc   ) {
		// check input
		HAWAII_ERROR_CONDITIONAL( mapU.empty()
		                       || mapV.empty(),
		                          "Transformation maps must not be empty." ) ;
		HAWAII_ERROR_CONDITIONAL( mapU.type() != CV_32FC1
		                       || mapV.type() != CV_32FC1,
		                          "Transformation map types must be 32-bit floating-point single-channel." ) ;
		HAWAII_ERROR_CONDITIONAL( mapU.size()
		                       != mapV.size(),
		                          "Transformation map sizes must be equal." ) ;
		
		// focal lengths already equal?
		const double fU = intrin.focalLengthU(),
		             fV = intrin.focalLengthV() ;
		if( fU == fV ) {
			intrinEqFoc = intrin ;
			mapUEqFoc   = mapU   ;
			mapVEqFoc   = mapV   ;
			return ;
		}
		
		// increase resolution of axis with shorter focal length
		const double scaleU = ( fV > fU ? fV / fU : 1.0 ),
		             scaleV = ( fU > fV ? fU / fV : 1.0 ) ;
		intrinEqFoc = intrin ;
		for( int col = 0 ; col < Intrinsic::cols ; ++col ) {
			intrinEqFoc( 0, col ) *= scaleU ;
			intrinEqFoc( 1, col ) *= scaleV ;
		}
		cv::resize( mapU, mapUEqFoc, cv::Size(), scaleU, scaleV, cv::INTER_CUBIC ) ;
		cv::resize( mapV, mapVEqFoc, cv::Size(), scaleU, scaleV, cv::INTER_CUBIC ) ;
		
	} // function "equalizeFocalLengths()"
	
} // anonymous namespace

namespace toast2 {
namespace stereo {

// generic remapping: initialize or update based on LUTs from output to input coordinates
Remapper::Remapper( const cv::Mat mapUArg,
                    const cv::Mat mapVArg ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( (  mapUArg.empty() && !mapVArg.empty() )
	                       || ( !mapUArg.empty() &&  mapVArg.empty() ),
	                          "Either none or both transformation maps must be empty." ) ;
	// wrap method below
	if( !mapUArg.empty()
	 && !mapVArg.empty() ) {
		this->update( mapUArg,
		              mapVArg ) ;
	}
}
void Remapper::update( const cv::Mat mapUArg,
                       const cv::Mat mapVArg ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( mapUArg.empty()
	                       || mapVArg.empty(),
	                          "Transformation maps must not be empty." ) ;
	HAWAII_ERROR_CONDITIONAL( mapUArg.type() != CV_32FC1
	                       || mapVArg.type() != CV_32FC1,
	                          "Transformation map types must be 32-bit floating-point single-channel." ) ;
	HAWAII_ERROR_CONDITIONAL( mapUArg.size()
	                       != mapVArg.size(),
	                          "Transformation map sizes must be equal." ) ;
	
	// convert to index-based representation
	cv::Mat mapPixArg, mapSubPixArg ;
	cv::convertMaps( mapUArg, mapVArg, mapPixArg, mapSubPixArg, CV_16SC2, false ) ;
	
	// get exclusive write access and replace previous maps
	boost::lock_guard< boost::shared_mutex > writeLock( this->mutex ) ;
	this->updateUnlocked( mapUArg, mapVArg, mapPixArg, mapSubPixArg ) ;
}
void Remapper::updateUnlocked( const cv::Mat mapUArg,
                               const cv::Mat mapVArg,
                               const cv::Mat mapPixArg,
                               const cv::Mat mapSubPixArg ) {
	// deep-copy u/v maps
	mapUArg.copyTo( this->mapU.writeCPU( mapUArg.size(), mapUArg.type() ) ) ;
	mapVArg.copyTo( this->mapV.writeCPU( mapVArg.size(), mapVArg.type() ) ) ;
	
	// shallow-copy index-based maps
	this->mapPix    = mapPixArg    ;
	this->mapSubPix = mapSubPixArg ;
}

// generic remapping: perform transformation on CPU
void Remapper::CPU( const cv::Mat  src,
                          cv::Mat& dst,
                          cv::Rect roi,
                    const int      cores ) const {
	
	// handle default ROI
	if( roi.x == 0 && roi.width  == 0
	 && roi.y == 0 && roi.height == 0 ) {
		roi.width  = this->mapPix.cols ;
		roi.height = this->mapPix.rows ;
	}
	
	// allocate result
	dst.create( roi.size(), src.type() ) ;
	
	// prevent writing, but allow other readers
	boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
	
	// check state
	HAWAII_ERROR_CONDITIONAL( this->mapPix.empty()
	                       || this->mapSubPix.empty(),
	                          "Transformation maps are not initialized." ) ;
	HAWAII_ERROR_CONDITIONAL( this->mapPix.size()
	                       != this->mapSubPix.size(),
	                          "Transformation map sizes must be equal." ) ;
	HAWAII_ERROR_CONDITIONAL( roi.x < 0 || ( roi.x + roi.width  ) > this->mapPix.cols
	                       || roi.y < 0 || ( roi.y + roi.height ) > this->mapPix.rows,
	                          "Region of interest exceeds LUT size."                  ) ;
	
	// partition LUTs and output
	const auto mapPixParts    = hawaii::partitionizeHoriz( this->mapPix(    roi ), cores ) ;
	const auto mapSubPixParts = hawaii::partitionizeHoriz( this->mapSubPix( roi ), cores ) ;
	const auto dstParts       = hawaii::partitionizeHoriz( dst,                    cores ) ;
	
	// remap each partition
	#pragma omp parallel for                \
		if( cores > 1 ) num_threads( cores ) \
		default( none )
	for( int core = 0 ; core < cores ; ++core ) {
		cv::remap( src, dstParts[ core ],
		           mapPixParts[ core ], mapSubPixParts[ core ],
		           cv::INTER_LINEAR, cv::BORDER_CONSTANT, 0    ) ;
	}
	
} // method "Remapper::CPU()"

// generic remapping: perform transformation on GPU
void Remapper::GPU( const cv::gpu::GpuMat  src,
                          cv::gpu::GpuMat& dst,
                          cv::Rect         roi,
                          cv::gpu::Stream& stream ) const {
	
	// check compile-time GPU support
	#ifndef HAWAII_NO_CUDA
		
		// handle default ROI
		if( roi.x == 0 && roi.width  == 0
		 && roi.y == 0 && roi.height == 0 ) {
			roi.width  = this->mapPix.cols ;
			roi.height = this->mapPix.rows ;
		}
		
		// allocate result
		hawaii::GPU::memoryPool( dst, roi.size(), src.type() ) ;
		
		// prevent writing, but allow other readers
		boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
		
		// check state
		HAWAII_ERROR_CONDITIONAL( this->mapU.empty()
		                       || this->mapV.empty(),
		                          "Transformation maps are not initialized." ) ;
		HAWAII_ERROR_CONDITIONAL( this->mapU.size()
		                       != this->mapV.size(),
		                          "Transformation map sizes must be equal." ) ;
		HAWAII_ERROR_CONDITIONAL( roi.x < 0 || ( roi.x + roi.width  ) > this->mapU.cols
		                       || roi.y < 0 || ( roi.y + roi.height ) > this->mapU.rows,
		                          "Region of interest exceeds LUT size."                ) ;
		
		// remap the image, specify stream in case copying is necessary
		cv::gpu::remap( src, dst,
		                this->mapU.readGPU( stream )( roi ),
		                this->mapV.readGPU( stream )( roi ),
		                cv::INTER_LINEAR, cv::BORDER_CONSTANT, 0,
		                stream                                   ) ;
		
	// cause run-time error, but avoid compile-time warnings
	#else
		HAWAII_ERROR_NO_CUDA ;
		(void)src    ;
		(void)dst    ;
		(void)roi    ;
		(void)stream ;
	#endif
	
} // method "Remapper::GPU()"

// generic remapping: get initialization state
bool Remapper::initialized() const {
	boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
	return ( !this->mapU.empty()
	      && !this->mapV.empty()
	      && !this->mapPix.empty()
	      && !this->mapSubPix.empty() ) ;
}

// generic remapping: get output size
cv::Size Remapper::sizeDst() const {
	boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
	HAWAII_ERROR_CONDITIONAL( !this->initialized(),
	                          "not yet initialized" ) ;
	return this->mapU.size() ;
}

// undistort: create un-initialized instance
Undistorter::Undistorter() {
}

// undistort: initialize or update based on distorted image/camera parameters
Undistorter::Undistorter( const Distortion distortion,
                          const Camera     cameraSrc,
                          const cv::Size   sizeSrc,
                          const bool       equalFocalLengths ) {
	// wrap method below
	this->update( distortion,
	              cameraSrc,
	              sizeSrc,
	              equalFocalLengths ) ;
}
void Undistorter::update( const Distortion distortion,
                          const Camera     cameraSrc,
                          const cv::Size   sizeSrc,
                          const bool       equalFocalLengths ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( isnan( distortion( 0 ) ),
	                          "invalid distortion parameters" ) ;
	HAWAII_ERROR_CONDITIONAL( isnan( cameraSrc( 0, 0 ) ),
	                          "invalid distorted camera matrix" ) ;
	HAWAII_ERROR_CONDITIONAL( sizeSrc.width  == 0
	                       || sizeSrc.height == 0,
	                          "invalid distorted image size" ) ;
	
	// find undistorted camera matrix to maximize field of view while avoiding any invalid pixels
	cv::Size sizeDst = sizeSrc ;
	const cv::Mat cameraDstTemp = cv::getOptimalNewCameraMatrix( cameraSrc,     // input: distorted camera matrix
	                                                             distortion,    // input: distortion coefficients
	                                                             sizeSrc,       // input: distorted image size
	                                                             0.0,           // parameter: 0.0 (avoid invalid pixels) ... 1.0 (include all source pixels)
	                                                             sizeDst,       // input (optional): undistorted image size, defaults to distorted size
	                                                             nullptr,       // result (optional): pointer to "cv::Rect" region of valid pixels
	                                                             false      ) ; // parameter: move principal point into undistorted image's center
	Camera cameraDstArg ;
	for( int row = 0 ; row < Camera::rows ; ++row ) {
	for( int col = 0 ; col < Camera::cols ; ++col ) {
		cameraDstArg( row, col ) = cameraDstTemp.at< double >( row, col ) ;
	} }
	
	// wrap method based on distorted and undistorted parameters below
	this->update( distortion,
	              cameraSrc,
	              cameraDstArg,
	              sizeDst,
	              equalFocalLengths ) ;
}

// initialize or update based on distorted/undistorted image/camera parameters
Undistorter::Undistorter( const Distortion distortion,
                          const Camera     cameraSrc,
                          const Camera     cameraDstArg,
                          const cv::Size   sizeDst,
                          const bool       equalFocalLengths ) {
	// wrap method below
	this->update( distortion,
	              cameraSrc,
	              cameraDstArg,
	              sizeDst,
	              equalFocalLengths ) ;
}
void Undistorter::update( const Distortion distortion,
                          const Camera     cameraSrc,
                          const Camera     cameraDstArg,
                          const cv::Size   sizeDst,
                          const bool       equalFocalLengths ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( isnan( distortion( 0 ) ),
	                          "invalid distortion parameters" ) ;
	HAWAII_ERROR_CONDITIONAL( isnan( cameraSrc( 0, 0 ) ),
	                          "invalid distorted camera matrix" ) ;
	HAWAII_ERROR_CONDITIONAL( isnan( cameraDstArg( 0, 0 ) ),
	                          "invalid undistorted camera matrix" ) ;
	HAWAII_ERROR_CONDITIONAL( sizeDst.width  == 0
	                       || sizeDst.height == 0,
	                          "invalid undistorted image size" ) ;
	
	// compute floating point LUTs
	cv::Mat mapUArg, mapVArg ;
	cv::initUndistortRectifyMap( cameraSrc,    // input: distorted camera matrix
	                             distortion,   // input: distortion coefficients
	                             cv::Mat(),    // input (optional): homography e.g. for rectification
	                             cameraDstArg, // input: undistorted camera matrix
	                             sizeDst,      // parameter: undistorted image size
	                             CV_32FC1,     // parameter: type of maps below
	                             mapUArg,      // result: U-coordinate if above is "CV_32FC1", pixel index if "CV_16SC2"
	                             mapVArg ) ;   // result: V-coordinate ",                      sub-pixel index "
	
	// wrap LUT-based method below
	this->update( mapUArg,
	              mapVArg,
	              cameraDstArg,
	              equalFocalLengths ) ;
}

// undistort: initialize or update based on LUTs and undistorted camera matrix
Undistorter::Undistorter( const cv::Mat mapUArg,
                          const cv::Mat mapVArg,
                          const Camera  cameraDstArg,
                          const bool    equalFocalLengths ) {
	// wrap method below
	this->update( mapUArg,
	              mapVArg,
	              cameraDstArg,
	              equalFocalLengths ) ;
}
void Undistorter::update( const cv::Mat mapUArg,
                          const cv::Mat mapVArg,
                          const Camera  cameraDstArg,
                          const bool    equalFocalLengths ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( mapUArg.empty()
	                       || mapVArg.empty(),
	                          "Transformation maps must not be empty." ) ;
	HAWAII_ERROR_CONDITIONAL( mapUArg.type() != CV_32FC1
	                       || mapVArg.type() != CV_32FC1,
	                          "Transformation map types must be 32-bit floating-point single-channel." ) ;
	HAWAII_ERROR_CONDITIONAL( mapUArg.size()
	                       != mapVArg.size(),
	                          "Transformation map sizes must be equal." ) ;
	HAWAII_ERROR_CONDITIONAL( isnan( cameraDstArg( 0, 0 ) ),
	                          "Invalid undistorted camera matrix" ) ;
	
	// equalize focal lengths
	Camera  cameraDstEqFoc ;
	cv::Mat mapUEqFoc,
	        mapVEqFoc      ;
	if( equalFocalLengths ) {
		equalizeFocalLengths( cameraDstArg,   mapUArg,   mapVArg,
		                      cameraDstEqFoc, mapUEqFoc, mapVEqFoc ) ;
	} else {
		cameraDstEqFoc = cameraDstArg ;
		mapUEqFoc      = mapUArg      ;
		mapVEqFoc      = mapVArg      ;
	}
	
	// convert to index-based representation
	cv::Mat mapPixEqFoc, mapSubPixEqFoc ;
	cv::convertMaps( mapUEqFoc, mapVEqFoc, mapPixEqFoc, mapSubPixEqFoc, CV_16SC2, false ) ;
	
	// get exclusive write access and replace previous maps
	boost::lock_guard< boost::shared_mutex > writeLock( this->mutex ) ;
	this->cameraDstM = cameraDstEqFoc ;
	this->updateUnlocked( mapUEqFoc, mapVEqFoc, mapPixEqFoc, mapSubPixEqFoc ) ;
}

// undistort: get output camera matrix
Camera Undistorter::cameraDst() const {
	boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
	HAWAII_ERROR_CONDITIONAL( !this->initialized(),
	                          "not yet initialized" ) ;
	return this->cameraDstM ;
}

// rectify: create un-initialized instance
Rectifier::Rectifier() {
}

// rectify: initialize or update based on relative position of cameras and distorted image/camera parameters
Rectifier::Rectifier( const cv::Matx33d rotation,
                      const cv::Vec3d   translation,
                      const Distortion  distortionLeft,
                      const Distortion  distortionRight,
                      const Camera      cameraSrcLeft,
                      const Camera      cameraSrcRight,
                      const cv::Size    sizeSrc,
                      const bool        equalFocalLengths ) {
	
	// wrap method below
	this->update( rotation,
	              translation,
	              distortionLeft,
	              distortionRight,
	              cameraSrcLeft,
	              cameraSrcRight,
	              sizeSrc,
	              equalFocalLengths ) ;
}
void Rectifier::update( const cv::Matx33d rotation,
                        const cv::Vec3d   translation,
                        const Distortion  distortionLeft,
                        const Distortion  distortionRight,
                        const Camera      cameraSrcLeft,
                        const Camera      cameraSrcRight,
                        const cv::Size    sizeSrc,
                        const bool        equalFocalLengths ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( isnan( distortionLeft(  0 ) )
	                       || isnan( distortionRight( 0 ) ),
	                          "invalid distortion parameters" ) ;
	HAWAII_ERROR_CONDITIONAL( isnan( cameraSrcLeft(  0, 0 ) )
	                       || isnan( cameraSrcRight( 0, 0 ) ),
	                          "Invalid distorted camera matrices" ) ;
	HAWAII_ERROR_CONDITIONAL( sizeSrc.width  == 0
	                       || sizeSrc.height == 0,
	                          "invalid distorted image size" ) ;
	
	// find rectification homographies
	cv::Size sizeDst = sizeSrc ;
	cv::Matx33d homographyLeft,
	            homographyRight ;
	cv::Matx34d projDstLeftArg,
	            projDstRightArg ;
	cv::Matx44d dispToDepth ;
	cv::stereoRectify( cameraSrcLeft,               // input: distorted left camera matrix
	                   distortionLeft,              // input: left distortion coefficients
	                   cameraSrcRight,              // input: distorted right camera matrix
	                   distortionRight,             // input: right distortion coefficients
	                   sizeSrc,                     // input: undistorted image size
	                   rotation,                    // input: rotation matrix between cameras
	                   translation,                 // input: translation vector between cameras
	                   homographyLeft,              // result: left rectification homography
	                   homographyRight,             // result: right rectification homography
	                   projDstLeftArg,              // result: left projection matrix after rectification
	                   projDstRightArg,             // result: right projection matrix after rectification
	                   dispToDepth,                 // result: helper for converting disparities to 3D points, e.g. via "cv::reprojectImageTo3D()"
	                   cv::CALIB_ZERO_DISPARITY,    // parameter: guarantee identical principal point after rectification
	                   0.0,                         // parameter: 0.0 (avoid invalid pixels) ... 1.0 (include all source pixels)
	                   sizeDst                  ) ; // parameter (optional): size of result image, defaults to size of input image
	
	// call the homography-based method below
	this->update( homographyLeft,
	              homographyRight,
	              distortionLeft,
	              distortionRight,
	              cameraSrcLeft,
	              cameraSrcRight,
	              projDstLeftArg,
	              projDstRightArg,
	              sizeDst,
	              equalFocalLengths ) ;
}

// rectify: initialize or update based on rectification homographies and distorted/rectified image/camera parameters
Rectifier::Rectifier( const cv::Matx33d homographyLeft,
                      const cv::Matx33d homographyRight,
                      const Distortion  distortionLeft,
                      const Distortion  distortionRight,
                      const Camera      cameraSrcLeft,
                      const Camera      cameraSrcRight,
                      const Projection  projDstLeftArg,
                      const Projection  projDstRightArg,
                      const cv::Size    sizeDst,
                      const bool        equalFocalLengths ) {
	
	// wrap method below
	this->update( homographyLeft,
	              homographyRight,
	              distortionLeft,
	              distortionRight,
	              cameraSrcLeft,
	              cameraSrcRight,
	              projDstLeftArg,
	              projDstRightArg,
	              sizeDst,
	              equalFocalLengths ) ;
}
void Rectifier::update( const cv::Matx33d homographyLeft,
                        const cv::Matx33d homographyRight,
                        const Distortion  distortionLeft,
                        const Distortion  distortionRight,
                        const Camera      cameraSrcLeft,
                        const Camera      cameraSrcRight,
                        const Projection  projDstLeftArg,
                        const Projection  projDstRightArg,
                        const cv::Size    sizeDst,
                        const bool        equalFocalLengths ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( isnan( distortionLeft(  0 ) )
	                       || isnan( distortionRight( 0 ) ),
	                          "invalid distortion parameters" ) ;
	HAWAII_ERROR_CONDITIONAL( isnan( cameraSrcLeft(  0, 0 ) )
	                       || isnan( cameraSrcRight( 0, 0 ) ),
	                          "Invalid distorted camera matrices" ) ;
	HAWAII_ERROR_CONDITIONAL( isnan( projDstLeftArg(  0, 0 ) )
	                       || isnan( projDstRightArg( 0, 0 ) ),
	                          "invalid rectified projection matrices" ) ;
	HAWAII_ERROR_CONDITIONAL( sizeDst.width  == 0
	                       || sizeDst.height == 0,
	                          "invalid rectified image size" ) ;
	
	// compute floating-point LUTs
	// developer note: see "Undistorter::update()" for explanation of arguments to "cv::initUndistortRectifyMap()"
	cv::Mat mapULeftArg,  mapVLeftArg,
	        mapURightArg, mapVRightArg ;
	cv::initUndistortRectifyMap( cameraSrcLeft,  distortionLeft,  homographyLeft,  projDstLeftArg,  sizeDst, CV_32FC1, mapULeftArg,  mapVLeftArg  ) ;
	cv::initUndistortRectifyMap( cameraSrcRight, distortionRight, homographyRight, projDstRightArg, sizeDst, CV_32FC1, mapURightArg, mapVRightArg ) ;
	
	// wrap LUT-based method below
	this->update( mapULeftArg,    mapURightArg,
	              mapVLeftArg,    mapVRightArg,
	              projDstLeftArg, projDstRightArg,
	              equalFocalLengths               ) ;
}

// rectify: initialize or update based on LUTs and rectified projection matrices
Rectifier::Rectifier( const cv::Mat    mapULeftArg,
                      const cv::Mat    mapURightArg,
                      const cv::Mat    mapVLeftArg,
                      const cv::Mat    mapVRightArg,
                      const Projection projDstLeftArg,
                      const Projection projDstRightArg,
                      const bool       equalFocalLengths ) {
	
	// wrap method below
	this->update( mapULeftArg,    mapURightArg,
	              mapVLeftArg,    mapVRightArg,
	              projDstLeftArg, projDstRightArg,
	              equalFocalLengths               ) ;
}
void Rectifier::update( const cv::Mat    mapULeftArg,
                        const cv::Mat    mapURightArg,
                        const cv::Mat    mapVLeftArg,
                        const cv::Mat    mapVRightArg,
                        const Projection projDstLeftArg,
                        const Projection projDstRightArg,
                        const bool       equalFocalLengths ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( mapULeftArg.empty() || mapURightArg.empty()
	                       || mapVLeftArg.empty() || mapVRightArg.empty(),
	                          "Transformation maps must not be empty."    ) ;
	HAWAII_ERROR_CONDITIONAL( mapULeftArg.type() != CV_32FC1 || mapURightArg.type() != CV_32FC1
	                       || mapVLeftArg.type() != CV_32FC1 || mapVRightArg.type() != CV_32FC1,
	                          "Transformation map types must be 32-bit floating-point single-channel." ) ;
	const cv::Size size = mapULeftArg.size() ;
	HAWAII_ERROR_CONDITIONAL( mapULeftArg.size() != size || mapURightArg.size() != size
	                       || mapVLeftArg.size() != size || mapVRightArg.size() != size,
	                          "Transformation map sizes must be equal."                 ) ;
	HAWAII_ERROR_CONDITIONAL( isnan( projDstLeftArg(  0, 0 ) )
	                       || isnan( projDstRightArg( 0, 0 ) ),
	                          "invalid rectified projection matrices" ) ;
	
	// equalize focal lengths
	Projection projDstLeftEqFoc, projDstRightEqFoc ;
	cv::Mat    mapULeftEqFoc,    mapURightEqFoc,
	           mapVLeftEqFoc,    mapVRightEqFoc    ;
	if( equalFocalLengths ) {
		equalizeFocalLengths( projDstLeftArg,    mapULeftArg,    mapVLeftArg,
		                      projDstLeftEqFoc,  mapULeftEqFoc,  mapVLeftEqFoc  ) ;
		equalizeFocalLengths( projDstRightArg,   mapURightArg,   mapVRightArg,
		                      projDstRightEqFoc, mapURightEqFoc, mapVRightEqFoc ) ;
	} else {
		projDstLeftEqFoc = projDstLeftArg ; projDstRightEqFoc = projDstRightArg ;
		mapULeftEqFoc    = mapULeftArg    ; mapURightEqFoc    = mapURightArg    ;
		mapVLeftEqFoc    = mapVLeftArg    ; mapVRightEqFoc    = mapVRightArg    ;
	}
	
	// convert to index-based representation
	cv::Mat mapPixLeftEqFoc,    mapPixRightEqFoc,
	        mapSubPixLeftEqFoc, mapSubPixRightEqFoc ;
	cv::convertMaps( mapULeftEqFoc,  mapVLeftEqFoc,  mapPixLeftEqFoc,  mapSubPixLeftEqFoc,  CV_16SC2, false ) ;
	cv::convertMaps( mapURightEqFoc, mapVRightEqFoc, mapPixRightEqFoc, mapSubPixRightEqFoc, CV_16SC2, false ) ;
	
	// get exclusive write access and replace previous maps
	boost::lock_guard< boost::shared_mutex > writeLock( this->mutex ) ;
	this->projectionDstLM = projDstLeftEqFoc  ;
	this->projectionDstRM = projDstRightEqFoc ;
	this->remapperL.updateUnlocked( mapULeftEqFoc,  mapVLeftEqFoc,  mapPixLeftEqFoc,  mapSubPixLeftEqFoc  ) ;
	this->remapperR.updateUnlocked( mapURightEqFoc, mapVRightEqFoc, mapPixRightEqFoc, mapSubPixRightEqFoc ) ;
}

// rectify: perform transformation of both images on CPU
void Rectifier::CPU( const cv::Mat  srcLeft,
                     const cv::Mat  srcRight,
                           cv::Mat& dstLeft,
                           cv::Mat& dstRight,
                           cv::Rect roi,
                     const int      cores    ) const {
	boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
	this->remapperL.CPU( srcLeft,  dstLeft,  roi, cores ) ;
	this->remapperR.CPU( srcRight, dstRight, roi, cores ) ;
}

// rectify: perform transformation of both images on GPU
void Rectifier::GPU( const cv::gpu::GpuMat  srcLeft,
                     const cv::gpu::GpuMat  srcRight,
                           cv::gpu::GpuMat& dstLeft,
                           cv::gpu::GpuMat& dstRight,
                           cv::Rect         roi,
                           cv::gpu::Stream& stream   ) const {
	boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
	this->remapperL.GPU( srcLeft,  dstLeft,  roi, stream ) ;
	this->remapperR.GPU( srcRight, dstRight, roi, stream ) ;
}
	
// rectify: get initialization state, output size and projection matrices
bool Rectifier::initialized() const {
	boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
	return ( this->remapperL.initialized()
	      && this->remapperR.initialized() ) ;
}
cv::Size Rectifier::sizeDst() const {
	boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
	HAWAII_ERROR_CONDITIONAL( !this->initialized(),
	                          "not yet initialized" ) ;
	HAWAII_ERROR_CONDITIONAL( this->remapperL.sizeDst()
	                       != this->remapperR.sizeDst(),
	                          "unequal remapper sizes"  ) ;
	return this->remapperL.sizeDst() ;
}

Projection Rectifier::projectionDstLeft() const {
	boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
	HAWAII_ERROR_CONDITIONAL( !this->initialized(),
	                          "not yet initialized" ) ;
	return this->projectionDstLM ;
}
Projection Rectifier::projectionDstRight() const {
	boost::shared_lock< boost::shared_mutex > readLock( this->mutex ) ;
	HAWAII_ERROR_CONDITIONAL( !this->initialized(),
	                          "not yet initialized" ) ;
	return this->projectionDstRM ;
}

} } // namespaces "toast2::stereo"
