// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// drone application for manual flight control
// ===========================================

#include "appManual.h"
#include "navdata.h"
#include "keyboard.h"
#include "gamepad.h"
#include "hawaii/common/ostream.h"
#include "hawaii/common/error.h"
#include <std_msgs/Float32MultiArray.h>
#include <std_msgs/UInt8MultiArray.h>
#include <std_msgs/Bool.h>
#include <std_srvs/Empty.h>
#include <SDL/SDL.h>
#include <SDL/SDL_mixer.h>
#include <boost/version.hpp>

// helpers only used here
namespace {
	
	// determine whether a "boost::thread" is still running
	bool isThreadRunning( boost::thread& thread ) {
		return ( thread.get_id() != boost::thread::id() ) &&
		#if( BOOST_VERSION >= 105300 )
			!thread.try_join_for( boost::chrono::microseconds( 100 ) ) ;
		#else
			!thread.timed_join( boost::posix_time::microseconds( 100 ) ) ;
		#endif
	}
	
	// normalize the raw values from a gamepad's analog sticks to [-1.0f 1.0f]
	class GamepadNormalizerStick {
		
		// internal state, initialization
		protected:
		float state ;
		public:
		GamepadNormalizerStick() :
			state( 0.0f ) {
		}
		
		// read without/after update
		public:
		operator float() { return this->state ; }
		float operator ()( const short stateRaw ) {
			
			// invert sign so that left and forward are positive (like in drone control commands)
			if( stateRaw < 0 ) { this->state = -(float)stateRaw / 32768.0f ; }
			else               { this->state = -(float)stateRaw / 32767.0f ; }
			return this->state ;
		}
	} ;
	
	// normalize the raw values from a gamepad's triggers to [0.0f 1.0f]
	class GamepadNormalizerTrigger {

		// internal state, initialization
		protected:
		float state ;
		public:
		GamepadNormalizerTrigger() :
			state( 0.0f ) {
		}

		// read without/after update
		public:
		operator float() { return this->state ; }
		float operator ()( const short stateRaw ) {
			
			// A trigger starts at 0 before being pressed for the first time, but later goes all the way back to -32768 when released.
			if( stateRaw != 0 || this->state != 0.0f ) {
				this->state = ( stateRaw + 32768.0f ) / 65535.0f ;
			}
			return this->state ;
		}
	} ;

} // anonymous namespace

// constructor
DroneAppManual::DroneAppManual() :
	
	// most recently received messages
	navdata(),
	
	// internal state: start in dis-engaged mode
	engaged( false ),
	
	// base connection to ROS
	initializerROS( "droneAppManual" ),
	spinner( 1 ), // dev. note: 1 thread is sufficient because this class's callbacks don't take long.
	nodeHandle(),
	
	// inbound communication with autonomous flight application
	subscriberAppCmdVel(  nodeHandle.subscribe< geometry_msgs::Twist >( "droneApp/cmd_vel", 1, &DroneAppManual::callbackAppCmdVel,  this ) ),
	subscriberAppTakeoff( nodeHandle.subscribe< std_msgs::Empty      >( "droneApp/takeoff", 1, &DroneAppManual::callbackAppTakeoff, this ) ),
	subscriberAppLand(    nodeHandle.subscribe< std_msgs::Empty      >( "droneApp/land",    1, &DroneAppManual::callbackAppLand,    this ) ),
	serverAppFlightAnim( nodeHandle.advertiseService( "droneApp/setFlightAnimation", &DroneAppManual::callbackAppFlightAnim, this ) ),
	timerAutonomous( false ),
	
	// outbound communication with autonomous flight application
	publisherAppEngaged(        nodeHandle.advertise< std_msgs::Bool              >( "droneApp/engaged",        1 ) ),
	publisherAppCmdVelDelta(    nodeHandle.advertise< geometry_msgs::Twist        >( "droneApp/cmd_velDelta",   1 ) ),
	publisherAppKeyboardKeys(   nodeHandle.advertise< std_msgs::UInt8MultiArray   >( "droneApp/keyboardKeys",   1 ) ),
	publisherAppGamepadButtons( nodeHandle.advertise< std_msgs::UInt8MultiArray   >( "droneApp/gamepadButtons", 1 ) ),
	publisherAppGamepadAnalog(  nodeHandle.advertise< std_msgs::Float32MultiArray >( "droneApp/gamepadAnalog",  1 ) ),
	
	// inbound communication with drone via "ardrone_autonomy"
	subscriberDroneNavdata( nodeHandle.subscribe( "ardrone/navdata", 1, &DroneAppManual::callbackDroneNavdata, this ) ),
	timerDrone( false ),
	
	// outbound communication with drone via "ardrone_autonomy"
	publisherDroneCmdVel(  nodeHandle.advertise< geometry_msgs::Twist >( "cmd_vel",         1 ) ),
	publisherDroneTakeoff( nodeHandle.advertise< std_msgs::Empty      >( "ardrone/takeoff", 1 ) ),
	publisherDroneLand(    nodeHandle.advertise< std_msgs::Empty      >( "ardrone/land",    1 ) ),
	publisherDroneReset(   nodeHandle.advertise< std_msgs::Empty      >( "ardrone/reset",   1 ) ),
	clientDroneFlatTrim(   nodeHandle.serviceClient< std_srvs::Empty              >( "ardrone/flattrim"           ) ),
	clientDroneIMURecalib( nodeHandle.serviceClient< std_srvs::Empty              >( "ardrone/imu_recalib"        ) ),
	clientDroneFlightAnim( nodeHandle.serviceClient< ardrone_autonomy::FlightAnim >( "ardrone/setflightanimation" ) ),
	
	// timeout detection
	timeoutAutonomous( true ),
	timeoutDrone(      true ) {
	
} // c'tor

// destructor
DroneAppManual::~DroneAppManual() {
	this->stop() ;
}

// run the drone application, wait for it to finish or stop it immediately
void DroneAppManual::start() {
	
	// start ROS processing
	this->spinner.start() ;
	
	// start main loop if not running already
	if( !isThreadRunning( this->loopThread ) ) {
		this->loopThread = boost::thread( &DroneAppManual::loopMethod, this ) ;
	}
}
void DroneAppManual::wait() {
	
	// wait for main loop to finish
	if( this->loopThread.joinable() ) {
		this->loopThread.join() ;
	}
	
	// stop ROS processing
	this->spinner.stop() ;
}
void DroneAppManual::stop() {
	
	// stop main loop immediately
	if( this->loopThread.joinable() ) {
		this->loopThread.interrupt() ;
		this->loopThread.join() ;
	}
	
	// stop ROS processing
	this->spinner.stop() ;
}

// communication with autonomous flight application: callbacks for subscribed topics and offered services
void DroneAppManual::callbackAppCmdVel( const geometry_msgs::Twist::ConstPtr& cmdVelPtr ) {
	
	// heartbeat for time-out detection
	this->timerAutonomous.tic() ;
	
	// buffer, merge and send command
	this->cmdVelAutonomous = *cmdVelPtr ;
	this->mergeAndPublishCmdVel() ;
}
void DroneAppManual::callbackAppTakeoff( const std_msgs::Empty::ConstPtr& takeoffPtr ) {
	
	// heartbeat for time-out detection
	this->timerAutonomous.tic() ;
	
	// launch ansynchronous thread to reset, calibrate and finally take off
	if( this->engaged
	 && !isThreadRunning( this->takeoffThread ) ) {
		(void)takeoffPtr ;
		this->takeoffThread = boost::thread( &DroneAppManual::takeoffMethod, this ) ;
	}
}
void DroneAppManual::callbackAppLand( const std_msgs::Empty::ConstPtr& landPtr ) {
	
	// heartbeat for time-out detection
	this->timerAutonomous.tic() ;
	
	// directly pass message through
	if( this->engaged ) {
		(void)landPtr ;
		this->publisherDroneLand.publish( std_msgs::Empty() ) ;
	}
}
bool DroneAppManual::callbackAppFlightAnim( ardrone_autonomy::FlightAnim::Request&  request,
                                            ardrone_autonomy::FlightAnim::Response& response ) {
	// heartbeat for time-out detection
	this->timerAutonomous.tic() ;
	
	// directly call service
	if( this->engaged ) {
		this->clientDroneFlightAnim.call( request, response ) ;
		if( !response.result ) { hawaii::cerr << "ERROR: failed to call flight animation service" << hawaii::endl ; }
	} else {
		response.result = false ;
	}
	return response.result ;
}

// communication with drone via "ardrone_autonomy": callbacks for subscribed topics
void DroneAppManual::callbackDroneNavdata( const ardrone_autonomy::Navdata::ConstPtr& navdataPtr ) {
	
	// heartbeat for time-out detection
	this->timerDrone.tic() ;
	
	// if battery level is below 20%, warn each time it decreases further
	if( navdataPtr->batteryPercent !=  0.0f
	 && navdataPtr->batteryPercent <= 20.0f
	 && navdataPtr->batteryPercent <  this->navdata.batteryPercent ) {
		hawaii::cerr << "WARNING: drone's battery level at " << navdataPtr->batteryPercent << "%" << hawaii::endl ;
	}
	
	// if battery SOC is below 10%, land
	if( navdataPtr->batteryPercent !=  0.0f
	 && navdataPtr->batteryPercent <= 10.0f
	 && navdataPtr->state != navdataStateDefault
	 && navdataPtr->state != navdataStateLanded  ) {
		hawaii::cerr << "ERROR: drone's battery level below 10%, we're going DOWN!!!" << hawaii::endl ;
		this->publisherDroneLand.publish( std_msgs::Empty() ) ;
	}
	
	// buffer latest message
	this->navdata = *navdataPtr ;
}

// combine "droneApp/cmd_vel" message with manual keyboard/gamepad inputs and send result to drone
void DroneAppManual::mergeAndPublishCmdVel() const {
	
	// evaluate autonomous commands only if engaged and if autonomous application has not timed-out
	geometry_msgs::Twist toBePublished = ( this->engaged && !this->timeoutAutonomous ) ?
	                                     this->cmdVelAutonomous :
	                                     geometry_msgs::Twist() ;
	// check given commands for "NAN"
	if( isnan( toBePublished.linear.x  ) ) { toBePublished.linear.x  = 0.0 ; hawaii::cerr << "WARNING: replaced \"NAN\" longitudinal flight command by zero before sending." << hawaii::endl ; }
	if( isnan( toBePublished.linear.y  ) ) { toBePublished.linear.y  = 0.0 ; hawaii::cerr << "WARNING: replaced \"NAN\" lateral flight command by zero before sending."      << hawaii::endl ; }
	if( isnan( toBePublished.linear.z  ) ) { toBePublished.linear.z  = 0.0 ; hawaii::cerr << "WARNING: replaced \"NAN\" vertical flight command by zero before sending."     << hawaii::endl ; }
	if( isnan( toBePublished.angular.z ) ) { toBePublished.angular.z = 0.0 ; hawaii::cerr << "WARNING: replaced \"NAN\" yaw flight command by zero before sending."          << hawaii::endl ; }
	
	// limit autonomous commands
	constexpr double limitAutoLinX     = 1.0,
	                 limitAutoLinY     = 1.0,
	                 limitAutoLinZUp   = 1.0,
	                 limitAutoLinZDown = 1.0,
	                 limitAutoAngZ     = 1.0 ;
	toBePublished.linear.x  = std::min( toBePublished.linear.x,   limitAutoLinX     ) ;
	toBePublished.linear.x  = std::max( toBePublished.linear.x,  -limitAutoLinX     ) ;
	toBePublished.linear.y  = std::min( toBePublished.linear.y,   limitAutoLinY     ) ;
	toBePublished.linear.y  = std::max( toBePublished.linear.y,  -limitAutoLinY     ) ;
	toBePublished.linear.z  = std::min( toBePublished.linear.z,   limitAutoLinZUp   ) ;
	toBePublished.linear.z  = std::max( toBePublished.linear.z,  -limitAutoLinZDown ) ;
	toBePublished.angular.z = std::min( toBePublished.angular.z,  limitAutoAngZ     ) ;
	toBePublished.angular.z = std::max( toBePublished.angular.z, -limitAutoAngZ     ) ;
	
	// scale manual control with difference between automated commands and manual control limits
	constexpr double limitManuLinX     = 1.0,
	                 limitManuLinY     = 1.0,
	                 limitManuLinZUp   = 1.0,
	                 limitManuLinZDown = 1.0,
	                 limitManuAngZ     = 1.0 ;
	if( this->cmdVelManual.linear.x  > 0.0 ) { if( toBePublished.linear.x  <  limitManuLinX     ) { toBePublished.linear.x  += (  limitManuLinX     - toBePublished.linear.x  ) * this->cmdVelManual.linear.x  ; } }
	else                                     { if( toBePublished.linear.x  > -limitManuLinX     ) { toBePublished.linear.x  -= ( -limitManuLinX     - toBePublished.linear.x  ) * this->cmdVelManual.linear.x  ; } }
	if( this->cmdVelManual.linear.y  > 0.0 ) { if( toBePublished.linear.y  <  limitManuLinY     ) { toBePublished.linear.y  += (  limitManuLinY     - toBePublished.linear.y  ) * this->cmdVelManual.linear.y  ; } }
	else                                     { if( toBePublished.linear.y  > -limitManuLinY     ) { toBePublished.linear.y  -= ( -limitManuLinY     - toBePublished.linear.y  ) * this->cmdVelManual.linear.y  ; } }
	if( this->cmdVelManual.linear.z  > 0.0 ) { if( toBePublished.linear.z  <  limitManuLinZUp   ) { toBePublished.linear.z  += (  limitManuLinZUp   - toBePublished.linear.z  ) * this->cmdVelManual.linear.z  ; } }
	else                                     { if( toBePublished.linear.z  > -limitManuLinZDown ) { toBePublished.linear.z  -= ( -limitManuLinZDown - toBePublished.linear.z  ) * this->cmdVelManual.linear.z  ; } }
	if( this->cmdVelManual.angular.z > 0.0 ) { if( toBePublished.angular.z <  limitManuAngZ     ) { toBePublished.angular.z += (  limitManuAngZ     - toBePublished.angular.z ) * this->cmdVelManual.angular.z ; } }
	else                                     { if( toBePublished.angular.z > -limitManuAngZ     ) { toBePublished.angular.z -= ( -limitManuAngZ     - toBePublished.angular.z ) * this->cmdVelManual.angular.z ; } }
	
	// before publishing, check if on-board stabilization may be activated by setting all values to zero
	constexpr double thresholdOnboardStabilization = 0.001 ;
	if( false
	 && fabs( toBePublished.linear.x  ) < thresholdOnboardStabilization
	 && fabs( toBePublished.linear.y  ) < thresholdOnboardStabilization
	 && fabs( toBePublished.linear.z  ) < thresholdOnboardStabilization
	 && fabs( toBePublished.angular.z ) < thresholdOnboardStabilization ) {
		toBePublished.linear.x  = 0.0 ;
		toBePublished.linear.y  = 0.0 ;
		toBePublished.linear.z  = 0.0 ;
		toBePublished.angular.x = 0.0 ;
		toBePublished.angular.y = 0.0 ;
		toBePublished.angular.z = 0.0 ;
	}
	this->publisherDroneCmdVel.publish( toBePublished ) ;
	
	// also publish difference between autonomous and actual control, but set insignificantly small values to zero
	constexpr double thresholdDeltaSignificant = 0.025 ;
	geometry_msgs::Twist delta ;
	delta.linear.x  = toBePublished.linear.x  - this->cmdVelAutonomous.linear.x  ;
	delta.linear.y  = toBePublished.linear.y  - this->cmdVelAutonomous.linear.y  ;
	delta.linear.z  = toBePublished.linear.z  - this->cmdVelAutonomous.linear.z  ;
	delta.angular.x = toBePublished.angular.x - this->cmdVelAutonomous.angular.x ;
	delta.angular.y = toBePublished.angular.y - this->cmdVelAutonomous.angular.y ;
	delta.angular.z = toBePublished.angular.z - this->cmdVelAutonomous.angular.z ;
	if( std::abs( delta.linear.x  ) < thresholdDeltaSignificant ) { delta.linear.x  = 0.0 ; }
	if( std::abs( delta.linear.y  ) < thresholdDeltaSignificant ) { delta.linear.y  = 0.0 ; }
	if( std::abs( delta.linear.z  ) < thresholdDeltaSignificant ) { delta.linear.z  = 0.0 ; }
	if( std::abs( delta.angular.x ) < thresholdDeltaSignificant ) { delta.angular.x = 0.0 ; }
	if( std::abs( delta.angular.y ) < thresholdDeltaSignificant ) { delta.angular.y = 0.0 ; }
	if( std::abs( delta.angular.z ) < thresholdDeltaSignificant ) { delta.angular.z = 0.0 ; }
	this->publisherAppCmdVelDelta.publish( delta ) ;
	
} // method "DroneAppManual::mergeAndPublishCmdVel()"

// reset (if necessary) and calibrate drone before taking off
void DroneAppManual::takeoffMethod() {
	
	// only do something if not flying already
	if( this->navdata.state == navdataStateDefault
	 || this->navdata.state == navdataStateInit
	 || this->navdata.state == navdataStateLanded  ) {
		
		// loop because we can't detect in advance if a reset is needed
		while( true ) {
			
			// abort if battery is too empty to start (limited at 20% by drone's firmware)
			if( this->navdata.batteryPercent <= 20.0f ) {
				hawaii::cerr << "ERROR: battery too low for takeoff" << hawaii::endl ;
				break ;
			}
			
			// perform IMU re-calibration and flat-trim
			hawaii::cout << "INFO: takeoff delayed for re-calibration..." << hawaii::endl ;
			std_srvs::Empty srvEmpty ;
			this->clientDroneIMURecalib.call( srvEmpty ) ;
			this->clientDroneFlatTrim.call(   srvEmpty ) ;
			boost::this_thread::sleep( boost::posix_time::milliseconds( 2000 ) ) ;
			
			// try to take off
			hawaii::cout << "INFO: takeoff triggered"  << hawaii::endl ;
			this->publisherDroneTakeoff.publish( std_msgs::Empty() ) ;
			boost::this_thread::sleep( boost::posix_time::milliseconds( 1000 ) ) ;
			
			// exit loop on success
			if( this->navdata.state != navdataStateDefault
			 && this->navdata.state != navdataStateLanded  ) {
				break ;
			}
			
			// reset, wait without alarming the watch dog
			hawaii::cout << "INFO: reset required before takeoff..." << hawaii::endl ;
			this->publisherDroneReset.publish( std_msgs::Empty() ) ;
			boost::this_thread::sleep( boost::posix_time::milliseconds( 2000 ) ) ;
		}
		
	} // if "not flying already?"
	
} // method "DroneAppManual::takeOffMethod()"

// main loop for querying manual input, checking time-outs etc.
void DroneAppManual::loopMethod() {
	
	// initialize required modules of "libSDL"
	HAWAII_ERROR_CONDITIONAL( SDL_Init( SDL_INIT_VIDEO
	                                  | SDL_INIT_AUDIO
	                                  | SDL_INIT_JOYSTICK ) != 0,
	                          "SDL failed to initialize."        ) ;
	
	// open a window to display key mapping illustration and catch keystrokes while in focus
	SDL_Surface* const imageKeyboardTemp = SDL_LoadBMP( "data/keyboard.bmp" ) ;
	HAWAII_ERROR_CONDITIONAL( imageKeyboardTemp == nullptr,
	                          "SDL failed to load image file." ) ;
	SDL_Surface* const screen = SDL_SetVideoMode( imageKeyboardTemp->w,
	                                              imageKeyboardTemp->h,
	                                              0, SDL_DOUBLEBUF     ) ;
	HAWAII_ERROR_CONDITIONAL( screen == nullptr,
	                          "SDL failed to set video mode." ) ;
	SDL_Surface* const imageKeyboard = SDL_DisplayFormat( imageKeyboardTemp ) ;
	HAWAII_ERROR_CONDITIONAL( imageKeyboard == nullptr,
	                          "SDL failed to convert image to display format." ) ;
	SDL_FreeSurface( imageKeyboardTemp ) ;
	HAWAII_ERROR_CONDITIONAL( SDL_BlitSurface( imageKeyboard, nullptr, screen, nullptr ) != 0,
	                          "SDL failed to blit image to screen."                           ) ;
	HAWAII_ERROR_CONDITIONAL( SDL_Flip( screen ) != 0,
	                          "SDL failed to flip screen buffers." ) ;
	SDL_WM_SetCaption( "droneApp", "droneApp" ) ;
	
	// initialize audio, load voice announcements
	HAWAII_ERROR_CONDITIONAL( Mix_OpenAudio( 44100, AUDIO_S16SYS, 2, 4096 ) != 0,
	                          "failed to open SDL audio"                         ) ;
	Mix_Chunk* const soundRestored = Mix_LoadWAV( "data/signalRestored.wav" ) ;
	Mix_Chunk* const soundLost     = Mix_LoadWAV( "data/signalLost.wav"     ) ;
	HAWAII_ERROR_CONDITIONAL( soundRestored == nullptr
	                       || soundLost     == nullptr,
	                          "failed to load sound file" ) ;
	
	// make keyboard key states available
	const uint8_t* const keyboardKeysSDL = SDL_GetKeyState( nullptr ) ;
	HAWAII_ERROR_CONDITIONAL( keyboardKeysSDL == nullptr,
	                          "SDL failed to get key states." ) ;
	
	// make first gamepad available
	SDL_Joystick* gamepadHandle = nullptr ;
	if( SDL_NumJoysticks() > 0 ) {
		gamepadHandle = SDL_JoystickOpen( 0 ) ;
		HAWAII_ERROR_CONDITIONAL( gamepadHandle == nullptr,
		                          "SDL failed to access gamepad." ) ;
		HAWAII_ERROR_CONDITIONAL( ( strcmp( SDL_JoystickName( 0 ), "Generic X-Box pad"       ) != 0
		                         && strcmp( SDL_JoystickName( 0 ), "Logitech Gamepad F710"   ) != 0
		                         && strcmp( SDL_JoystickName( 0 ), "Logitech Gamepad F510"   ) != 0
		                         && strcmp( SDL_JoystickName( 0 ), "Logitech Gamepad F310"   ) != 0
		                         && strcmp( SDL_JoystickName( 0 ), "Microsoft X-Box 360 pad" ) != 0 )
		                       || SDL_JoystickNumAxes(    gamepadHandle ) != 6
		                       || SDL_JoystickNumBalls(   gamepadHandle ) != 0
		                       || SDL_JoystickNumButtons( gamepadHandle ) != 11 // developer note: I only found ten.
		                       || SDL_JoystickNumHats(    gamepadHandle ) != 1,
		                          "Logitech F{3,5,7}10 or Microsoft Xbox gamepad in XInput mode required" ) ;
	}
	
	// helpers for interpreting keyboard "pulse width modulation" and gamepad analog sticks/triggers
	KeySmoother keySmootherLinX( 0.05f ),
	            keySmootherLinY( 0.05f ),
	            keySmootherLinZ( 0.05f ),
	            keySmootherAngZ( 0.05f ) ;
	GamepadNormalizerStick   gamepadNormalizerStickLeftHoriz,
	                         gamepadNormalizerStickLeftVerti,
	                         gamepadNormalizerStickRightHoriz,
	                         gamepadNormalizerStickRightVerti ;
	GamepadNormalizerTrigger gamepadNormalizerTriggerLeft,
	                         gamepadNormalizerTriggerRight    ;
	
	// helper to land before shutting down
	bool shutdownRequest = false ;
	
	// for catching a "boost::thread_interrupted"
	try {
		
		// loop to be exited via "break"
		while( true ) {
			
			// update and publish keyboard key states
			SDL_PumpEvents() ;
			std_msgs::UInt8MultiArray keyboardKeys ;
			keyboardKeys.data.resize( SDLK_LAST, 0 ) ;
			keyboardKeys.layout.dim.push_back( std_msgs::MultiArrayDimension() ) ;
			keyboardKeys.layout.dim[ 0 ].label = "keys" ;
			keyboardKeys.layout.dim[ 0 ].size   = SDLK_LAST ;
			keyboardKeys.layout.dim[ 0 ].stride = SDLK_LAST ;
			std::copy( keyboardKeysSDL, keyboardKeysSDL + SDLK_LAST, keyboardKeys.data.begin() ) ;
			this->publisherAppKeyboardKeys.publish( keyboardKeys ) ;
			
			// update and publish gamepad states
			std_msgs::UInt8MultiArray gamepadButtons ;
			gamepadButtons.data.resize( 12, 0 ) ;
			gamepadButtons.layout.dim.push_back( std_msgs::MultiArrayDimension() ) ;
			gamepadButtons.layout.dim[ 0 ].label = "gamepad buttons" ;
			gamepadButtons.layout.dim[ 0 ].size   = 12 ;
			gamepadButtons.layout.dim[ 0 ].stride = 12 ;
			std_msgs::Float32MultiArray gamepadAnalog ;
			gamepadAnalog.data.resize( 6, 0.0f ) ;
			gamepadAnalog.layout.dim.push_back( std_msgs::MultiArrayDimension() ) ;
			gamepadAnalog.layout.dim[ 0 ].label = "gamepad analog" ;
			gamepadAnalog.layout.dim[ 0 ].size   = 6 ;
			gamepadAnalog.layout.dim[ 0 ].stride = 6 ;
			if( gamepadHandle != nullptr ) {
				SDL_JoystickUpdate() ;
				gamepadButtons.data[ gamepadButtonA           ] =                                   SDL_JoystickGetButton( gamepadHandle,  0 )   ;
				gamepadButtons.data[ gamepadButtonB           ] =                                   SDL_JoystickGetButton( gamepadHandle,  1 )   ;
				gamepadButtons.data[ gamepadButtonX           ] =                                   SDL_JoystickGetButton( gamepadHandle,  2 )   ;
				gamepadButtons.data[ gamepadButtonY           ] =                                   SDL_JoystickGetButton( gamepadHandle,  3 )   ;
				gamepadButtons.data[ gamepadButtonL           ] =                                   SDL_JoystickGetButton( gamepadHandle,  4 )   ;
				gamepadButtons.data[ gamepadButtonR           ] =                                   SDL_JoystickGetButton( gamepadHandle,  5 )   ;
				gamepadButtons.data[ gamepadButtonBack        ] =                                   SDL_JoystickGetButton( gamepadHandle,  6 )   ;
				gamepadButtons.data[ gamepadButtonStart       ] =                                   SDL_JoystickGetButton( gamepadHandle,  7 )   ;
				gamepadButtons.data[ gamepadButtonUNKNOWN     ] =                                   SDL_JoystickGetButton( gamepadHandle,  8 )   ;
				gamepadButtons.data[ gamepadStickLeftPressed  ] =                                   SDL_JoystickGetButton( gamepadHandle,  9 )   ;
				gamepadButtons.data[ gamepadStickRightPressed ] =                                   SDL_JoystickGetButton( gamepadHandle, 10 )   ;
				gamepadButtons.data[ gamepadDigitalCross      ] =                                   SDL_JoystickGetHat(    gamepadHandle,  0 )   ;
				gamepadAnalog.data[  gamepadStickLeftHoriz    ] = gamepadNormalizerStickLeftHoriz(  SDL_JoystickGetAxis(   gamepadHandle,  0 ) ) ;
				gamepadAnalog.data[  gamepadStickLeftVerti    ] = gamepadNormalizerStickLeftVerti(  SDL_JoystickGetAxis(   gamepadHandle,  1 ) ) ;
				gamepadAnalog.data[  gamepadStickRightHoriz   ] = gamepadNormalizerStickRightHoriz( SDL_JoystickGetAxis(   gamepadHandle,  3 ) ) ;
				gamepadAnalog.data[  gamepadStickRightVerti   ] = gamepadNormalizerStickRightVerti( SDL_JoystickGetAxis(   gamepadHandle,  4 ) ) ;
				gamepadAnalog.data[  gamepadTriggerLeft       ] = gamepadNormalizerTriggerLeft(     SDL_JoystickGetAxis(   gamepadHandle,  2 ) ) ;
				gamepadAnalog.data[  gamepadTriggerRight      ] = gamepadNormalizerTriggerRight(    SDL_JoystickGetAxis(   gamepadHandle,  5 ) ) ;
				this->publisherAppGamepadButtons.publish( gamepadButtons ) ;
				this->publisherAppGamepadAnalog.publish(  gamepadAnalog  ) ;
			}
			
			// engage or dis-engage?
			     if( keyboardKeys.data[   SDLK_SPACE         ]
			      || gamepadButtons.data[ gamepadButtonRed   ] ) {
				this->engaged = false ;
				hawaii::cout << "INFO: dis-engaged" << hawaii::endl ;
			}
			else if( keyboardKeys.data[   SDLK_b             ]
			      || gamepadButtons.data[ gamepadButtonGreen ] ) {
				this->engaged = true  ;
				hawaii::cout << "INFO: ENGAGED"     << hawaii::endl ;
			}
			std_msgs::Bool engagedMsg ;
			engagedMsg.data = this->engaged ;
			this->publisherAppEngaged.publish( engagedMsg ) ;
			
			// land of take off?
			     if( keyboardKeys.data[   SDLK_g             ]
			      || gamepadButtons.data[ gamepadButtonBack  ] ) {
			    this->publisherDroneLand.publish( std_msgs::Empty() ) ;
			    this->engaged = false ;
				hawaii::cout << "INFO: landing triggered and dis-engaged" << hawaii::endl ;
			}
			else if( keyboardKeys.data[   SDLK_t             ]
				  || gamepadButtons.data[ gamepadButtonStart ] ) {
				if( !isThreadRunning( this->takeoffThread ) ) {
					this->takeoffThread = boost::thread( &DroneAppManual::takeoffMethod, this ) ;
				}
				hawaii::cout << "INFO: takeoff triggered"  << hawaii::endl ;
			}
			
			// create "continuous" values from keystrokes 
			keySmootherLinX( keyboardKeys.data[ SDLK_u ],    // forward or...
			                 keyboardKeys.data[ SDLK_j ] ) ; // ...backward
			keySmootherLinY( keyboardKeys.data[ SDLK_h ],    // strafe left or...
			                 keyboardKeys.data[ SDLK_k ] ) ; // ...right
			keySmootherLinZ( keyboardKeys.data[ SDLK_o ],    // up or...
			                 keyboardKeys.data[ SDLK_l ] ) ; // ...down
			keySmootherAngZ( keyboardKeys.data[ SDLK_y ]     // turn left (on both English...
			              || keyboardKeys.data[ SDLK_z ],    // ...and German keyboards) or...
			                 keyboardKeys.data[ SDLK_i ] ) ; // ...right
			
			// add keyboard and gamepad, send commands to drone
			this->cmdVelManual.linear.x  = std::min( 1.0f, std::max( -1.0f, keySmootherLinX + gamepadAnalog.data[ gamepadStickLeftVerti  ] ) ) ;
			this->cmdVelManual.linear.y  = std::min( 1.0f, std::max( -1.0f, keySmootherLinY + gamepadAnalog.data[ gamepadStickLeftHoriz  ] ) ) ;
			this->cmdVelManual.linear.z  = std::min( 1.0f, std::max( -1.0f, keySmootherLinZ + gamepadAnalog.data[ gamepadStickRightVerti ] ) ) ;
			this->cmdVelManual.angular.z = std::min( 1.0f, std::max( -1.0f, keySmootherAngZ + gamepadAnalog.data[ gamepadStickRightHoriz ] ) ) ;
			this->mergeAndPublishCmdVel() ;

			// quit on [Esc] or [Ctrl]+[C]
			if( keyboardKeys.data[ SDLK_ESCAPE ]
			 || ( keyboardKeys.data[ SDLK_c ]
			   && ( keyboardKeys.data[ SDLK_LCTRL ]
			     || keyboardKeys.data[ SDLK_RCTRL ] ) ) ) {
				shutdownRequest = true ;
				hawaii::cout << "INFO: shut down requested..." << hawaii::endl ;
			}
			if( shutdownRequest ) {
				
				// exit loop only if landed (or crashed)
				if( this->navdata.state == navdataStateDefault
				 || this->navdata.state == navdataStateInit
				 || this->navdata.state == navdataStateLanded ) {
					hawaii::cout << "INFO: shutting down" << hawaii::endl ;
					break ;
				}
				
				// if necessary, trigger landing before shutting down
				else if( navdata.state != navdataStateLanding ) {
					hawaii::cout << "INFO: landing triggered before shutdown" << hawaii::endl ;
					this->publisherDroneLand.publish( std_msgs::Empty() ) ;
				}
			}
			
			// estimate connection state with drone by putting the time since the last received message in a 
			// hysteresis, print console output and play sounds if changed
			if( this->timerDrone.isRunning() ) {
				if( this->timeoutDrone == false ) {
					if( this->timerDrone.toc() > 500.0 ) {
						this->timeoutDrone = true ;
						hawaii::cerr << "WARNING: connection to drone LOST" << hawaii::endl ;
						HAWAII_ERROR_CONDITIONAL( Mix_PlayChannel( -1, soundLost, 0 ) < 0,
						                          "failed to play sound file"             ) ;
					}
				} else {
					if( this->timerDrone.toc() < 200.0 ) {
						this->timeoutDrone = false ;
						hawaii::cout << "INFO: connection to drone restored" << hawaii::endl ;
						HAWAII_ERROR_CONDITIONAL( Mix_PlayChannel( -1, soundRestored, 0 ) < 0,
						                          "failed to play sound file"                 ) ;
					}
				}
			}
			
			// estimate connection state with autonomous flight application by putting the time since the last received 
			// message in a hysteresis, print console output (but don't play any sounds) if changed
			if( this->timerAutonomous.isRunning() ) {
				if( this->timeoutAutonomous == false ) {
					if( this->timerAutonomous.toc() > 1000.0 ) {
						this->timeoutAutonomous = true ;
						hawaii::cerr << "WARNING: connection to autonomous application LOST" << hawaii::endl ;
					}
				} else {
					if( this->timerAutonomous.toc() < 400.0 ) {
						this->timeoutAutonomous = false ;
						hawaii::cout << "INFO: connection to autonomous application restored" << hawaii::endl ;
					}
				}
			}
			
			// loop at (a little less than) 50 Hz, allow interruption
			boost::this_thread::sleep( boost::posix_time::milliseconds( 20 ) ) ;
			
		} // while "loop"
		
	// let sounds finish playing, shut down SDL before returning
	} catch( const boost::thread_interrupted& ) {}
	while( Mix_Playing( -1 ) != 0 ) {
		boost::this_thread::sleep( boost::posix_time::milliseconds( 100 ) ) ;
	}
	Mix_FreeChunk( soundRestored ) ;
	Mix_FreeChunk( soundLost     ) ;
	Mix_CloseAudio() ;
	SDL_JoystickClose( gamepadHandle ) ;
	SDL_FreeSurface( imageKeyboard ) ;
	SDL_Quit() ;
	
} // method "DroneAppManual::loopMethod()"
