// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// find minimum or maximum of a variable number of arguments with different yet compatible types
// =============================================================================================

#pragma once

#include <utility>
#include <type_traits>

namespace hawaii {

// two implementations to choose from
#if( 1 )
	
	// adapted from "stackoverflow.com/a/7472115", added rvalue references and forwarding
	template< typename T > T min( T&& t ) { return std::forward< T >( t ) ; } // dev. note: allow template termination
	template< typename T > T max( T&& t ) { return std::forward< T >( t ) ; } // "
	template< typename T1, typename T2, typename ... Tn, typename C = typename std::common_type< T1, T2, Tn ... >::type >
	C min( T1&& t1, T2&& t2, Tn&& ... tn ) {
		return std::forward< C >( hawaii::min( std::forward< C >( t1 <= t2 ? t1 : t2 ),
		                                       std::forward< C >( tn                ) ... ) ) ;
	}
	template< typename T1, typename T2, typename ... Tn, typename C = typename std::common_type< T1, T2, Tn ... >::type >
	C max( T1&& t1, T2&& t2, Tn&& ... tn ) {
		return std::forward< C >( hawaii::max( std::forward< C >( t1 >= t2 ? t1 : t2 ),
		                                       std::forward< C >( tn                ) ... ) ) ;
	}
	
#else
	
	// own divide-and-conquer implementation after reading "stackoverflow.com/questions/5484930"
	// user note: This is more than 1000 times slower than the code above; forwarding did not improve this. I decided to 
	//            keep it only as a reference for implementing much more complex parallel reductions.
	template< size_t n, typename C >
	C minImpl( const typename std::initializer_list< C >::iterator& iter ) {
		if( n == 1 ) { return *iter ; } // dev. note: no partial function
		const C minA = minImpl< ( n + 1 ) / 2, C >( iter                 ),
		        minB = minImpl<   n       / 2, C >( iter + ( n + 1 ) / 2 ) ;
		return minA <= minB ? minA : minB ;
	}
	template< size_t n, typename C >
	C maxImpl( const typename std::initializer_list< C >::iterator& iter ) {
		if( n == 1 ) { return *iter ; } // "
		const C maxA = maxImpl< ( n + 1 ) / 2, C >( iter                 ),
		        maxB = maxImpl<   n       / 2, C >( iter + ( n + 1 ) / 2 ) ;
		return maxA >= maxB ? maxA : maxB ;
	}
	template< typename ... Tn, typename C = typename std::common_type< Tn ... >::type >
	C min( Tn&& ... tn ) { return minImpl< sizeof ... ( tn ), C >( std::initializer_list< C >( { tn ... } ).begin() ) ; }
	template< typename ... Tn, typename C = typename std::common_type< Tn ... >::type >
	C max( Tn&& ... tn ) { return maxImpl< sizeof ... ( tn ), C >( std::initializer_list< C >( { tn ... } ).begin() ) ; }
	
#endif

} // namespace "hawaii"
