// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// wrapper to include and enhance OpenCV 2/3.x.x's GPU/CUDA module
// ===============================================================

#include "hawaii/GPU/openCV.h"
#include "hawaii/GPU/CUDA.h"
#include "hawaii/GPU/pool.h"
#include "hawaii/common/error.h"
#ifndef HAWAII_NO_CUDA
	#if CV_MAJOR_VERSION == 3
		#define __OPENCV_BUILD
		#define HAVE_CUDA
		#include <opencv2/core/private.cuda.hpp>
		#undef __OPENCV_BUILD
		#undef HAVE_CUDA
	#endif
#endif

// helpers only used here
namespace {
	
	// reproduce members and sub-classes of "cv::{cuda,gpu}::Stream" which are encapsulated using the pimpl pattern  
	// developer note: keep in sync with the definition and implementation in "modules/core/{include/opencv2/core/
	//                 cuda.hpp,src/cuda_stream.cpp}" (OpenCV 3) or "modules/gpu/{include/opencv2/gpu/gpu.hpp,src/
	//                 cudastream.cpp}" (OpenCV 2)
	#ifndef HAWAII_NO_CUDA
		#if CV_MAJOR_VERSION == 3
			struct StackAllocatorAccessible : public cv::gpu::GpuMat::Allocator {
				cudaStream_t stream_ ;
				void* memStack_ ;
				size_t alignment_ ;
			} ;
			struct StreamAccessible {
				struct Impl {
					cudaStream_t stream ;
					cv::Ptr< StackAllocatorAccessible > stackAllocator_ ;
				} ;
				cv::Ptr< Impl > impl_ ;
			} ;
		#elif CV_MAJOR_VERSION == 2
			struct StreamAccessible {
				struct Impl {
					cudaStream_t stream ;
					int ref_counter ;
				} ;
				Impl* impl ;
			} ;
		#else
			#error "unsupported version of OpenCV"
		#endif
	#endif
}

namespace hawaii {
namespace GPU {

// unified functions for asynchronous copying between CPU and GPU
void copyAsync( const cv::gpu::CudaMem         srcCudaMem,
                      cv::gpu::GpuMat&         dstGpuMat,
                      cv::gpu::Stream&         stream,
                      hawaii::GPU::MemoryPool& pool       ) {
	if( dstGpuMat.size() != srcCudaMem.size()
	 || dstGpuMat.type() != srcCudaMem.type() ) {
		pool( dstGpuMat, srcCudaMem.size(), srcCudaMem.type() ) ;
	}
	#if CV_MAJOR_VERSION == 3
		dstGpuMat.upload( srcCudaMem, stream ) ;
	#elif CV_MAJOR_VERSION == 2
		stream.enqueueUpload( srcCudaMem, dstGpuMat ) ;
	#else
		#error "unsupported version of OpenCV"
	#endif
}
void copyAsync( const cv::gpu::GpuMat          srcGpuMat,
                      cv::gpu::CudaMem&        dstCudaMem,
                      cv::gpu::Stream&         stream,
                      hawaii::GPU::MemoryPool& pool       ) {
	if( dstCudaMem.size() != srcGpuMat.size()
	 || dstCudaMem.type() != srcGpuMat.type() ) {
		pool( dstCudaMem, srcGpuMat.size(), srcGpuMat.type() ) ;
	}
	#if CV_MAJOR_VERSION == 3
		srcGpuMat.download( dstCudaMem, stream ) ;
	#elif CV_MAJOR_VERSION == 2
		stream.enqueueDownload( srcGpuMat, dstCudaMem ) ;
	#else
		#error "unsupported version of OpenCV"
	#endif
}
void copyAsync( const cv::gpu::GpuMat          srcGpuMat,
                      cv::gpu::GpuMat&         dstGpuMat,
                      cv::gpu::Stream&         stream,
                      hawaii::GPU::MemoryPool& pool      ) {
	if( dstGpuMat.size() != srcGpuMat.size()
	 || dstGpuMat.type() != srcGpuMat.type() ) {
		pool( dstGpuMat, srcGpuMat.size(), srcGpuMat.type() ) ;
	}
	#if CV_MAJOR_VERSION == 3
		srcGpuMat.copyTo( dstGpuMat, stream ) ;
	#elif CV_MAJOR_VERSION == 2
		stream.enqueueCopy( srcGpuMat, dstGpuMat ) ;
	#else
		#error "unsupported version of OpenCV"
	#endif
}

// unified functions for asynchronous batch assignment
void assignAsync(       cv::gpu::GpuMat  dstGpuMat,
                  const cv::Scalar       value,
                        cv::gpu::Stream& stream    ) {
	#if CV_MAJOR_VERSION == 3
		dstGpuMat.setTo( value, stream ) ;
	#elif CV_MAJOR_VERSION == 2
		stream.enqueueMemSet( dstGpuMat, value ) ;
	#else
		#error "unsupported version of OpenCV"
	#endif
}
void assignAsync(       cv::gpu::GpuMat  dstGpuMat,
                  const cv::gpu::GpuMat  mask,
                  const cv::Scalar       value,
                        cv::gpu::Stream& stream    ) {
	#if CV_MAJOR_VERSION == 3
		dstGpuMat.setTo( value, mask, stream ) ;
	#elif CV_MAJOR_VERSION == 2
		stream.enqueueMemSet( dstGpuMat, value, mask ) ;
	#else
		#error "unsupported version of OpenCV"
	#endif
}

// create an instance of "cv::{gpu,cuda}::Stream" which does not implicitly synchronize with the default stream
cv::gpu::Stream streamNonBlockingFactory() {
	
	// "cv::gpu::Stream::Stream()" or "HAWAII_ERROR_NO_CUDA" throw if compile-time GPU support is missing in OpenCV or 
	// libHawaii respectively.
	cv::gpu::Stream stream ;
	HAWAII_ERROR_NO_CUDA ;
	#ifndef HAWAII_NO_CUDA
		
		// suppress warning
		#pragma GCC diagnostic push
		#pragma GCC diagnostic ignored "-Wstrict-aliasing"
		StreamAccessible* streamAccessible = (StreamAccessible*)&stream ;
		#pragma GCC diagnostic pop
		
		// replace the actual "cudaStream_t" which OpenCV created via "cudaStreamCreate(...)" by one created via 
		// "cudaStreamCreateWithFlags( ..., cudaStreamNonBlocking )"
		#if CV_MAJOR_VERSION == 3
			HAWAII_ERROR_CHECK_CUDA( cudaStreamDestroy( streamAccessible->impl_->stream ) ) ;
			HAWAII_ERROR_CHECK_CUDA( cudaStreamCreateWithFlags( &streamAccessible->impl_->stream, cudaStreamNonBlocking ) ) ;
			streamAccessible->impl_->stackAllocator_->stream_ = streamAccessible->impl_->stream ;
		#elif CV_MAJOR_VERSION == 2
			HAWAII_ERROR_CHECK_CUDA( cudaStreamDestroy( streamAccessible->impl->stream ) ) ;
			HAWAII_ERROR_CHECK_CUDA( cudaStreamCreateWithFlags( &streamAccessible->impl->stream, cudaStreamNonBlocking ) ) ;
		#else
			#error "unsupported version of OpenCV"
		#endif
		
	#endif
	return stream ;
}

} } // namespaces "hawaii::GPU"
